from __future__ import annotations
from typing import Optional
from jsonclasses.config import Config
from inflection import pluralize


class DBConf:

    def __init__(self: DBConf,
                 cls: type,
                 config: Config,
                 collection_name: Optional[str],
                 camelize_db_keys: Optional[bool]) -> None:
        self._cls = cls
        self._config = config
        self._collection_name = (collection_name or
                                 pluralize(cls.__name__).lower())
        self._camelize_db_keys = (camelize_db_keys if camelize_db_keys is not
                                  None else True)

    @property
    def config(self: DBConf) -> Config:
        return self._config

    @property
    def collection_name(self: DBConf) -> str:
        return self._collection_name

    @property
    def camelize_db_keys(self: DBConf) -> bool:
        return self._camelize_db_keys



from __future__ import annotations
from typing import Any, Optional, TypeVar, cast, TYPE_CHECKING
from datetime import date
from jsonclasses.types import Types
from jsonclasses.field_definition import FieldStorage, FieldType
from jsonclasses.types_resolver import TypesResolver
from jsonclasses.mark_graph import MarkGraph
from inflection import underscore, camelize
from .utils import (ref_field_key, ref_field_keys, ref_db_field_key,
                    ref_db_field_keys)
from .coder import Coder
from .dbconf import DBConf
if TYPE_CHECKING:
    from .pymongo_object import PymongoObject
    T = TypeVar('T', bound=PymongoObject)


class Decoder(Coder):

    def decode_list(self,
                    value: list[Any],
                    cls: type[T],
                    types: Types,
                    graph: MarkGraph) -> Optional[list[Any]]:
        if value is None:
            return None
        item_types = TypesResolver().resolve_types(
            types.definition.raw_item_types, cls.definition.config)
        new_cls = item_types.definition.instance_types
        return ([self.decode_item(value=item, cls=new_cls,
                                  types=item_types,
                                  graph=graph)
                for item in value])

    def decode_dict(self,
                    value: dict[str, Any],
                    cls: type[T],
                    types: Types,
                    graph: MarkGraph) -> Optional[dict[str, Any]]:
        if value is None:
            return None
        config: DBConf = cls.dbconf
        if types.definition.field_storage == FieldStorage.FOREIGN_KEY:
            return None
        if types.definition.field_storage == FieldStorage.LOCAL_KEY:
            return ({underscore(k) if config.camelize_db_keys else k: str(v)
                    for k, v in value.items()})
        else:
            item_types = TypesResolver().resolve_types(
                types.definition.raw_item_types)
            return ({underscore(k) if config.camelize_db_keys else k:
                    self.decode_item(value=v, cls=cls, types=item_types,
                                     graph=graph)
                    for k, v in value.items()})

    def decode_shape(self,
                     value: dict[str, Any],
                     cls: type[T],
                     types: Types,
                     graph: MarkGraph) -> dict[str, Any]:
        config: DBConf = cls.dbconf
        shape_types = cast(dict[str, Any], types.definition.shape_types)
        retval = {}
        for k, item_types in shape_types.items():
            retval[k] = self.decode_item(
                value=value[(camelize(k, False)
                             if config.camelize_db_keys else k)],
                cls=cls,
                types=item_types,
                graph=graph)
        return retval

    def decode_item(self,
                    value: Any,
                    cls: type[T],
                    types: Types,
                    graph: MarkGraph) -> Any:
        if value is None:
            return value
        if types.definition.field_type == FieldType.DATE:
            return date.fromisoformat(value.isoformat()[:10])
        elif types.definition.field_type == FieldType.ENUM:
            if isinstance(types.definition.enum_class, str):
                t = TypesResolver().resolve_types(types.definition.enum_class,
                                                  cls.definition.config)
                enum_cls = cast(type, t.definition.enum_class)
                return enum_cls(value)
            else:
                enum_cls = cast(type, types.definition.enum_class)
                return enum_cls(value)
        elif types.definition.field_type == FieldType.LIST:
            return self.decode_list(value=value, cls=cls, types=types,
                                    graph=graph)
        elif types.definition.field_type == FieldType.DICT:
            return self.decode_dict(value=value, cls=cls, types=types,
                                    graph=graph)
        elif types.definition.field_type == FieldType.SHAPE:
            return self.decode_shape(value=value, cls=cls, types=types,
                                     graph=graph)
        elif types.definition.field_type == FieldType.INSTANCE:
            return self.decode_instance(value=value, cls=cls, types=types,
                                        graph=graph)
        else:
            return value

    def decode_instance(self,
                        value: dict[str, Any],
                        cls: type[T],
                        types: Types,
                        graph: MarkGraph) -> Any:
        inst_id = str(value.get('_id'))
        dest = graph.getp(cls, inst_id)
        exist = True
        if dest is None:
            dest = cls()
            exist = False
        for field in cls.definition.fields:
            if cls.dbconf.camelize_db_keys:
                key = camelize(field.name, False)
            else:
                key = field.name
            if self.is_id_field(field):
                if not exist:
                    setattr(dest, field.name, inst_id)
                    graph.put(dest)
            elif self.is_foreign_key_storage(field):
                if value.get(key) is not None:
                    if isinstance(value.get(key), list):
                        t = TypesResolver().resolve_types(
                            field.definition.raw_item_types,
                            cls.definition.config)
                        new_cls = t.definition.instance_types
                        setattr(dest, field.name,
                                self.decode_list(
                                    value[key], new_cls, field.types, graph))
                    else:
                        t = TypesResolver().resolve_types(
                            field.definition.instance_types,
                            cls.definition.config)
                        new_cls = t.definition.instance_types
                        setattr(dest, field.name,
                                self.decode_instance(
                                    cast(dict[str, Any], value.get(key)),
                                    new_cls, field.types, graph))
            elif self.is_local_key_reference_field(field):
                if value.get(key) is not None:
                    t = TypesResolver().resolve_types(
                        field.definition.instance_types,
                        cls.definition.config)
                    new_cls = t.definition.instance_types
                    inst = self.decode_item(
                        value=value.get(key), types=field.types, cls=new_cls,
                        graph=graph)
                    setattr(dest, field.name, inst)
                ref_id = value.get(ref_db_field_key(field.name, cls=cls))
                setattr(dest, ref_field_key(field.name), str(ref_id))
            elif self.is_local_keys_reference_field(field):
                if value.get(key) is not None:
                    t = TypesResolver().resolve_types(
                        field.definition.raw_item_types,
                        cls.definition.config)
                    new_cls = t.definition.instance_types
                    setattr(dest, field.name,
                            self.decode_list(
                                value[key], new_cls, field.types, graph))

                saved_keys = value.get(ref_db_field_keys(field.name, cls))
                setattr(dest, ref_field_keys(field.name),
                        [str(k) for k in saved_keys])
            elif self.is_instance_field(field):
                t = TypesResolver().resolve_types(
                    field.definition.instance_types,
                    cls.definition.config)
                new_cls = t.definition.instance_types
                setattr(dest, field.name, self.decode_item(
                    value=value.get(key), types=field.types, cls=new_cls,
                    graph=graph))
            else:
                if not exist:
                    setattr(
                        dest,
                        field.name,
                        self.decode_item(value=value.get(key),
                                         types=field.types,
                                         cls=cls, graph=graph))
        return dest

    def apply_initial_status(self, root: T,
                             graph: Optional[MarkGraph] = None) -> None:
        if graph is None:
            graph = MarkGraph()
        if graph.get(root) is not None:
            return
        graph.put(root)
        for field in root.__class__.definition.fields:
            if self.is_instance_field(field):
                if getattr(root, field.name) is not None:
                    self.apply_initial_status(getattr(root, field.name), graph)
            elif self.is_list_instance_field(field, root.__class__):
                if getattr(root, field.name) is not None:
                    for item in getattr(root, field.name):
                        self.apply_initial_status(item, graph)
        root._mark_unmodified()

    def decode_root(self,
                    root: dict[str, Any],
                    cls: type[T],
                    graph: Optional[MarkGraph] = None) -> T:
        if graph is None:
            graph = MarkGraph()
        types = Types().instanceof(cls)
        decoded = self.decode_instance(root, cls, types, graph)
        self.apply_initial_status(decoded)
        return decoded

    def decode_root_list(self,
                         root_list: list[dict[str, Any]],
                         cls: type[T],
                         graph: Optional[MarkGraph] = None) -> list[T]:
        if graph is None:
            graph = MarkGraph()
        types = Types().instanceof(cls)
        results: list[T] = []
        for root in root_list:
            decoded = self.decode_instance(root, cls, types, graph)
            results.append(decoded)
        for result in results:
            self.apply_initial_status(result)
        return results



# def _include(self, key_path: str):
#     field = next(field for field in fields(self)
#                  if field.field_name == key_path)
#     fd = field.field_types.field_description
#     if (fd.field_type == FieldType.INSTANCE and
#             fd.field_storage == FieldStorage.LOCAL_KEY):
#         ref_id = getattr(self, ref_field_key(field.field_name))
#         if ref_id is not None:
#             Cls = cast(Type[MongoObject], resolve_types(
#                     fd.instance_types,
#                     graph_sibling=self.__class__
#                 ).field_description.instance_types)
#             setattr(self, field.field_name, Cls.find_by_id(ref_id))
#     elif (fd.field_type == FieldType.INSTANCE and
#             fd.field_storage == FieldStorage.FOREIGN_KEY):
#         foreign_key_name = ref_db_field_key(cast(str, fd.foreign_key),
#                                             self.__class__)
#         Cls = cast(Type[MongoObject], resolve_types(
#             fd.instance_types,
#             graph_sibling=self.__class__).field_description.instance_types)
#         setattr(self, field.field_name, Cls.find_one(
#             {foreign_key_name: ObjectId(self.id)}))
#     elif (fd.field_type == FieldType.LIST and
#             fd.field_storage == FieldStorage.LOCAL_KEY):
#         ref_ids = getattr(self, ref_field_keys(field.field_name))
#         if ref_ids is not None:
#             item_types = resolve_types(
#                 fd.list_item_types, self.__class__)
#             Cls = cast(Type[MongoObject], resolve_types(
#                 item_types.field_description.instance_types,
#                 self.__class__))
#             setattr(self, field.field_name, Cls.find_many(
#                 {'_id': {'$in': [ObjectId(id) for id in ref_ids]}}))
#     elif (fd.field_type == FieldType.LIST and
#             fd.field_storage == FieldStorage.FOREIGN_KEY):
#         if fd.use_join_table:
#             decoder = Decoder()
#             self_class = self.__class__
#             other_class = decoder.list_instance_type(
#                 field, self.__class__)
#             join_table_name = decoder.join_table_name(
#                 self_class,
#                 field.field_name,
#                 other_class,
#                 cast(str, fd.foreign_key)
#             )
#             jt_collection = self_class.db().get_collection(join_table_name)
#             cursor = jt_collection.aggregate([
#                 {
#                     '$match': {
#                         ref_db_field_key(self_class.__name__, self_class):
#                             ObjectId(self.id)
#                     }
#                 },
#                 {
#                     '$lookup': {
#                         'from': other_class.collection().name,
#                         'localField': ref_db_field_key(
#                                 other_class.__name__,
#                                 other_class),
#                         'foreignField': '_id',
#                         'as': 'result'
#                     }
#                 },
#                 {
#                     '$project': {
#                         'result': {'$arrayElemAt': ['$result', 0]}
#                     }
#                 },
#                 {
#                     '$replaceRoot': {
#                         'newRoot': '$result'
#                     }
#                 }
#             ])
#             results = [decoder.decode_root(
#                 doc, other_class) for doc in cursor]
#             setattr(self, field.field_name, results)
#         else:
#             foreign_key_name = ref_db_field_key(
#                 cast(str, fd.foreign_key), self.__class__)
#             item_types = resolve_types(
#                 fd.list_item_types, self.__class__)
#             Cls = cast(Type[MongoObject], resolve_types(
#                 item_types.field_description.instance_types,
#                 self.__class__).field_description.instance_types)
#             setattr(self, field.field_name, Cls.find_many(
#                 {foreign_key_name: ObjectId(self.id)}))
#     else:
#         pass

# def include(self: T, *args: str) -> T:
#     for arg in args:
#         self._include(arg)
#     return self



class DatabaseNotConnectedException(Exception):
    """This exception is raised when collection is accessed however database is
    not connected.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)



from __future__ import annotations
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import pymongo, BaseObject


@pymongo
@jsonclass(class_graph='linked')
class LinkedCourse(BaseObject):
    name: str
    students: list[LinkedStudent] = types.listof('LinkedStudent') \
                                         .linkedthru('courses')


@pymongo
@jsonclass(class_graph='linked')
class LinkedStudent(BaseObject):
    name: str
    courses: list[LinkedCourse] = types.listof('LinkedCourse') \
                                       .linkedthru('students')



from __future__ import annotations
from enum import Enum
from jsonclasses import jsonclass, jsonenum
from jsonclasses_pymongo import pymongo, BaseObject


@jsonenum(class_graph='simple')
class Gender(Enum):
    MALE = 1
    FEMALE = 2


@pymongo
@jsonclass(class_graph='simple')
class SimpleArtist(BaseObject):
    name: str
    gender: Gender



from __future__ import annotations
from typing import Union, TYPE_CHECKING
from inflection import camelize
from jsonclasses.field_definition import FieldType
if TYPE_CHECKING:
    from .pymongo_object import PymongoObject


def ref_key(key: str, cls: type[PymongoObject]) -> tuple[str, str]:
    field_name = key + '_id'
    if cls.dbconf.camelize_db_keys:
        db_field_name = camelize(field_name, False)
    else:
        db_field_name = field_name
    return (field_name, db_field_name)


def ref_field_key(key: str) -> str:
    return key + '_id'


def ref_db_field_key(key: str, cls: type[PymongoObject]) -> str:
    field_name = ref_field_key(key)
    if cls.dbconf.camelize_db_keys:
        db_field_name = camelize(field_name, False)
    else:
        db_field_name = field_name
    return db_field_name


def ref_field_keys(key: str) -> str:
    return key + '_ids'


def ref_db_field_keys(key: str, cls: type[PymongoObject]) -> str:
    field_name = ref_field_keys(key)
    if cls.dbconf.camelize_db_keys:
        db_field_name = camelize(field_name, False)
    else:
        db_field_name = field_name
    return db_field_name


def btype_from_ftype(field_type: FieldType) -> Union[str, list[str]]:
    if field_type == FieldType.STR:
        return 'string'
    elif field_type == FieldType.INT:
        return ['int', 'long', 'decimal']
    elif field_type == FieldType.FLOAT:
        return 'double'
    elif field_type == FieldType.BOOL:
        return 'bool'
    elif field_type == FieldType.DATE:
        return 'date'
    elif field_type == FieldType.DATETIME:
        return 'date'
    else:
        return []



from __future__ import annotations
from typing import TypeVar, Any, Union, cast
from re import search
from bson.objectid import ObjectId
from jsonclasses.jsonclass_field import JSONClassField
from pymongo.errors import DuplicateKeyError
from inflection import camelize, underscore
from pymongo.collection import Collection
from jsonclasses.field_definition import FieldStorage, FieldType
from jsonclasses.exceptions import UniqueConstraintException
from jsonclasses.types_resolver import TypesResolver
from jsonclasses.exceptions import DeletionDeniedException
from .pymongo_object import PymongoObject
from .query import BaseQuery, ExistQuery, IterateQuery, ListQuery, SingleQuery, IDQuery
from .encoder import Encoder
from .connection import Connection
from .utils import btype_from_ftype, ref_db_field_key
from .coder import Coder


T = TypeVar('T', bound=PymongoObject)


def find(cls: type[T], **kwargs: Any) -> ListQuery[T]:
    return ListQuery(cls=cls, filter=kwargs)


def one(cls: type[T], **kwargs: Any) -> SingleQuery[T]:
    return SingleQuery(cls=cls, filter=kwargs)


def pymongo_id(cls: type[T], id: Union[str, ObjectId]) -> IDQuery[T]:
    return IDQuery(cls=cls, id=id)


def linked(cls: type[T]) -> BaseQuery[T]:
    return BaseQuery(cls=cls)


def exist(cls: type[T], **kwargs: Any) -> ExistQuery[T]:
    return ExistQuery(cls=cls, filter=kwargs)


def iterate(cls: type[T], **kwargs: Any) -> IterateQuery[T]:
    return IterateQuery(cls=cls, filter=kwargs)


def _database_write(self: T) -> None:
    try:
        Encoder().encode_root(self).execute()
    except DuplicateKeyError as exception:
        result = search('index: (.+?)_1', exception._message)
        assert result is not None
        db_key = result.group(1)
        pt_key = db_key
        json_key = db_key
        if self.__class__.dbconf.camelize_db_keys:
            pt_key = underscore(db_key)
            json_key = pt_key
        if self.__class__.definition.config.camelize_json_keys:
            json_key = camelize(pt_key, False)
        raise UniqueConstraintException(
                getattr(self, pt_key), json_key) from None


def _orm_delete(self: T, no_raise: bool = False) -> None:
    # deny test
    for field in self.__class__.definition.deny_fields:
        if field.definition.field_storage == FieldStorage.LOCAL_KEY:
            key = self.__class__.definition.config.key_transformer(field)
            if hasattr(self, key) and getattr(self, key) is not None:
                if no_raise:
                    return
                else:
                    raise DeletionDeniedException()
        elif field.definition.field_storage == FieldStorage.FOREIGN_KEY:
            r = TypesResolver()
            if field.definition.field_type == FieldType.LIST:
                t = r.resolve_types(field.definition.raw_item_types,
                                    self.__class__.definition.config)
            else:
                t = r.resolve_types(field.definition.instance_types,
                                    self.__class__.definition.config)
            types = t
            oc = cast(type[PymongoObject], types.definition.instance_types)
            f = cast(JSONClassField, field.foreign_field)
            if field.definition.use_join_table:
                jtname = Coder().join_table_name(
                    self.__class__,
                    field.name,
                    oc,
                    f.name)
                coll = Connection.from_class(self.__class__).collection(jtname)
                key = ref_db_field_key(self.__class__.__name__, self.__class__)
                e = coll.count_documents({key: ObjectId(self._id)}, limit=1)
                exist = e > 0
                if exist:
                    if no_raise:
                        return
                    else:
                        raise DeletionDeniedException()
            else:
                key = oc.definition.config.key_transformer(f)
                exist = oc.exist(**{key: ObjectId(self._id)}).exec()
                if exist:
                    if no_raise:
                        return
                    else:
                        raise DeletionDeniedException()

    collection = Connection.get_collection(self.__class__)
    collection.delete_one({'_id': ObjectId(self._id)})

    # delete chain - nullify
    for field in self.__class__.definition.nullify_fields:
        if field.definition.field_storage == FieldStorage.FOREIGN_KEY:
            r = TypesResolver()
            if field.definition.field_type == FieldType.LIST:
                t = r.resolve_types(field.definition.raw_item_types,
                                    self.__class__.definition.config)
            else:
                t = r.resolve_types(field.definition.instance_types,
                                    self.__class__.definition.config)
            types = t
            oc = cast(type[PymongoObject], types.definition.instance_types)
            f = cast(JSONClassField, field.foreign_field)
            if field.definition.use_join_table:
                jtname = Coder().join_table_name(
                    self.__class__,
                    field.name,
                    oc,
                    f.name)
                coll = Connection.from_class(self.__class__).collection(jtname)
                key = ref_db_field_key(self.__class__.__name__, self.__class__)
                coll.delete_many({key: ObjectId(self._id)})
            else:
                key = oc.definition.config.key_transformer(f)
                for o in oc.iterate(**{key: ObjectId(self._id)}).exec():
                    setattr(o, f.name, None)
                    setattr(o, key, None)
                    o.save(skip_validation=True)

    # delete chain - cascade
    for field in self.__class__.definition.cascade_fields:
        r = TypesResolver()
        if field.definition.field_type == FieldType.LIST:
            t = r.resolve_types(field.definition.raw_item_types,
                                self.__class__.definition.config)
        else:
            t = r.resolve_types(field.definition.instance_types,
                                self.__class__.definition.config)
        types = t
        types = t
        oc = cast(type[PymongoObject], types.definition.instance_types)
        f = cast(JSONClassField, field.foreign_field)
        if field.definition.field_storage == FieldStorage.LOCAL_KEY:
            key = self.__class__.definition.config.key_transformer(field)
            if getattr(self, key) is not None:
                item = oc.id(getattr(self, key)).optional.exec()
                if item is not None:
                    item._orm_delete(no_raise=True)
        elif field.definition.field_storage == FieldStorage.FOREIGN_KEY:
            if field.definition.use_join_table:
                jtname = Coder().join_table_name(
                    self.__class__,
                    field.name,
                    oc,
                    f.name)
                coll = Connection.from_class(self.__class__).collection(jtname)
                key = ref_db_field_key(self.__class__.__name__, self.__class__)
                other_key = ref_db_field_key(oc.__name__, oc)
                for rel in coll.find({key: ObjectId(self._id)}):
                    other_id = rel[other_key]
                    item = oc.id(other_id).optional.exec()
                    if item is not None:
                        item._orm_delete(no_raise=True)
                coll.delete_many({key: ObjectId(self._id)})
            else:
                key = oc.definition.config.key_transformer(f)
                for o in oc.iterate(**{key: ObjectId(self._id)}).exec():
                    o._orm_delete(no_raise=True)

    setattr(self, '_is_deleted', True)


def _orm_restore(self: T) -> None:
    pass


def pymongofy(class_: type) -> PymongoObject:
    # do not install methods for subclasses
    if hasattr(class_, '__is_pymongo__'):
        return cast(PymongoObject, class_)
    # type marks
    setattr(class_, '__is_pymongo__', True)
    # public methods
    class_.find = classmethod(find)
    class_.one = classmethod(one)
    class_.id = classmethod(pymongo_id)
    class_.linked = classmethod(linked)
    class_.exist = classmethod(exist)
    class_.iterate = classmethod(iterate)
    # protected methods
    class_._database_write = _database_write
    class_._orm_delete = _orm_delete
    class_._orm_restore = _orm_restore
    connection = Connection.from_class(class_)
    if class_.definition.config.abstract:
        return class_
    def callback(coll: Collection):
        info = coll.index_information()
        for field in class_.definition.fields:
            name = field.name
            if class_.dbconf.camelize_db_keys:
                name = camelize(field.name, False)
            index = field.definition.index
            unique = field.definition.unique
            required = field.definition.required
            index_name = f'{name}_1'
            ftype = field.definition.field_type
            if unique:
                if required:
                    coll.create_index(name, name=index_name, unique=True)
                else:
                    coll.create_index(
                        name, name=index_name, unique=True,
                        partialFilterExpression={
                            name: {'$type': btype_from_ftype(ftype)}
                        })
            elif index:
                if required:
                    coll.create_index(name, name=index_name)
                else:
                    coll.create_index(
                        name, name=index_name,
                        partialFilterExpression={
                            name: {'$type': btype_from_ftype(ftype)}
                        })
            else:
                if index_name in info.keys():
                    coll.drop_index(index_name)
    connection.add_connected_callback(class_.dbconf.collection_name, callback)
    return class_



from __future__ import annotations
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import pymongo, BaseObject


@pymongo
@jsonclass(class_graph='linked')
class LinkedNote(BaseObject):
    name: str
    notebook: LinkedNotebook = types.instanceof('LinkedNotebook').linkto


@pymongo
@jsonclass(class_graph='linked')
class LinkedNotebook(BaseObject):
    name: str
    notes: list[LinkedNote] = types.listof('LinkedNote').linkedby('notebook') \
                                   .deny


@pymongo
@jsonclass(class_graph='linked')
class LinkedRNote(BaseObject):
    name: str
    notebook: LinkedRNotebook = types.instanceof('LinkedRNotebook').linkto.deny


@pymongo
@jsonclass(class_graph='linked')
class LinkedRNotebook(BaseObject):
    name: str
    notes: list[LinkedRNote] = types.listof('LinkedRNote').linkedby('notebook')



"""This module contains queries."""
from __future__ import annotations
from typing import (Iterator, Union, TypeVar, Generator, Optional, Any,
                    Generic, NamedTuple, cast)
from bson import ObjectId
from inflection import camelize
from pymongo.cursor import Cursor
from jsonclasses.field_definition import FieldStorage, FieldType
from jsonclasses.types_resolver import TypesResolver
from jsonclasses.mark_graph import MarkGraph
from jsonclasses.exceptions import ObjectNotFoundException
from .coder import Coder
from .decoder import Decoder
from .connection import Connection
from .pymongo_object import PymongoObject
from .utils import ref_db_field_key
T = TypeVar('T', bound=PymongoObject)
U = TypeVar('U', bound='BaseQuery')
V = TypeVar('V', bound='BaseListQuery')


class Subquery(NamedTuple):
    name: str
    query: Optional[BaseQuery]


class BaseQuery(Generic[T]):
    """Base query is the base class of queries.
    """

    def __init__(self: U, cls: type[T]) -> None:
        self._cls = cls
        self.subqueries: list[Subquery] = []

    def include(self: U, name: str, query: Optional[BaseQuery] = None) -> U:
        self.subqueries.append(Subquery(name, query))
        return self

    def _build_aggregate_pipeline(self: U) -> list[dict[str, Any]]:
        cls = cast(type[PymongoObject], self._cls)
        result: list[dict[str, Any]] = []
        for subquery in self.subqueries:
            fname = subquery.name
            field = cls.definition.field_named(fname)
            tr = TypesResolver()
            if field.definition.field_type == FieldType.LIST:
                types = tr.resolve_types(field.definition.raw_item_types,
                                         cls.definition.config)
            else:
                types = tr.resolve_types(field.definition.instance_types,
                                         cls.definition.config)
            it = cast(type[PymongoObject], types.definition.instance_types)
            if field.definition.field_storage == FieldStorage.LOCAL_KEY:
                key = ref_db_field_key(fname, cls)
                if subquery.query is None:
                    result.append({
                        '$lookup': {
                            'from': it.dbconf.collection_name,
                            'localField': key,
                            'foreignField': '_id',
                            'as': fname
                        }
                    })
                else:
                    subpipeline = subquery.query._build_aggregate_pipeline()
                    subpipeline.insert(0, {
                        '$match': {
                            '$expr': {
                                '$and': [{'$eq': ['$_id', '$$'+key]}]
                            }
                        }
                    })
                    result.append({
                        '$lookup': {
                            'from': it.dbconf.collection_name,
                            'as': fname,
                            'let': {key: '$' + key},
                            'pipeline': subpipeline
                        }
                    })
                result.append({
                    '$unwind': '$' + fname
                })
            elif field.definition.field_storage == FieldStorage.FOREIGN_KEY:
                if field.definition.field_type == FieldType.INSTANCE:
                    fk = cast(str, field.definition.foreign_key)
                    if subquery.query is None:
                        result.append({
                            '$lookup': {
                                'from': it.dbconf.collection_name,
                                'localField': '_id',
                                'foreignField': ref_db_field_key(fk, it),
                                'as': fname
                            }
                        })
                    else:
                        fk = cast(str, field.definition.foreign_key)
                        key = ref_db_field_key(fk, it)
                        subp = subquery.query._build_aggregate_pipeline()
                        subp.insert(0, {
                            '$match': {
                                '$expr': {
                                    '$and': [{'$eq': ['$'+key, '$$'+key]}]
                                }
                            }
                        })
                        result.append({
                            '$lookup': {
                                'from': it.dbconf.collection_name,
                                'as': fname,
                                'let': {key: '$_id'},
                                'pipeline': subp
                            }
                        })
                    result.append({
                        '$unwind': '$' + fname
                    })
                elif field.definition.field_type == FieldType.LIST:
                    if subquery.query is not None:
                        subpipeline = subquery.query \
                                                ._build_aggregate_pipeline()
                    else:
                        subpipeline = []
                    has_match = False
                    matcher = None
                    for item in subpipeline:
                        if item.get('$match'):
                            has_match = True
                            matcher = item.get('$match')
                            break
                    if field.definition.use_join_table:
                        coder = Coder()
                        jt_name = coder.join_table_name(cls, field.name,
                                                        it, field.foreign_field.name)
                        this_key = ref_db_field_key(cls.__name__, cls)
                        that_key = ref_db_field_key(it.__name__, it)
                        pipeline: list[Any] = []
                        if matcher is None:
                            matcher = {}
                        if not matcher.get('$expr'):
                            matcher['$expr'] = {}
                        if not matcher['$expr'].get('$and'):
                            matcher['$expr']['$and'] = []
                        matcher['$expr']['$and'].append({
                            '$eq': ['$_id', '$$' + that_key]
                        })
                        if not has_match:
                            subpipeline.insert(0, {'$match': matcher})
                        outer_lookup = {'$lookup': {
                            'from': jt_name,
                            'as': field.name,
                            'let': {this_key: '$_id'},
                            'pipeline': pipeline
                        }}
                        match = {'$match': {
                            '$expr': {
                                '$and': [
                                    {'$eq': ['$'+this_key, '$$'+this_key]}
                                ]
                            }
                        }}
                        lookup = {'$lookup': {
                            'from': it.dbconf.collection_name,
                            'as': field.name,
                            'let': {that_key: '$'+that_key},
                            'pipeline': subpipeline
                        }}
                        unwind = {
                            '$unwind': '$' + field.name
                        }
                        replace = {
                            '$replaceRoot': {'newRoot': '$'+field.name}
                        }
                        pipeline.append(match)
                        pipeline.append(lookup)
                        pipeline.append(unwind)
                        pipeline.append(replace)
                        result.append(outer_lookup)
                    else:
                        fk = cast(str, field.definition.foreign_key)
                        key = ref_db_field_key(fk, it)
                        item = {
                            '$lookup': {
                                'from': it.dbconf.collection_name,
                                'as': fname,
                                'let': {key: '$_id'},
                                'pipeline': subpipeline
                            }
                        }
                        if matcher is None:
                            matcher = {}
                        if not matcher.get('$expr'):
                            matcher['$expr'] = {}
                        if not matcher['$expr'].get('$and'):
                            matcher['$expr']['$and'] = []
                        matcher['$expr']['$and'].append({
                            '$eq': ['$' + key, '$$' + key]
                        })
                        if not has_match:
                            subpipeline.insert(0, {'$match': matcher})
                        result.append(item)
        return result


class BaseListQuery(BaseQuery[T]):
    """Base list query is the base class of list queries.
    """

    def __init__(self: V,
                 cls: type[T],
                 filter: Optional[dict[str, Any]] = None) -> None:
        super().__init__(cls)
        self._match: Optional[dict[str, Any]] = None
        self._sort: Optional[list[tuple[str, int]]] = None
        self._skip: Optional[int] = None
        self._limit: Optional[int] = None
        self._use_pick: bool = False
        self._pick: Optional[dict[str, Any]] = None
        self._use_omit: bool = False
        self._omit: Optional[dict[str, Any]] = None
        if filter is not None:
            self._set_matcher(filter)

    def _set_matcher(self: V, matcher: dict[str, Any]) -> None:
        cls = cast(PymongoObject, self._cls)
        if cls.definition.primary_field is not None:
            pf_name: Optional[str] = cls.definition.primary_field.name
        else:
            pf_name = None
        cls.definition._camelized_reference_names
        result: dict[str, Any] = {}
        for key, value in matcher.items():
            if key == pf_name:
                new_value = ObjectId(value)
            elif key in cls.definition._camelized_reference_names:
                new_value = ObjectId(value)
            elif key in cls.definition._reference_names:
                new_value = ObjectId(value)
            else:
                new_value = value
            if cls.dbconf.camelize_db_keys:
                result[camelize(key, False)] = new_value
        self._match = result

    def order(self: V, field: str, sort: Optional[int] = None) -> V:
        if self._sort is None:
            self._sort = []
        self._sort.append((field, sort or 1))
        return self

    def skip(self: V, n: int) -> V:
        self._skip = n
        return self

    def limit(self: V, n: int) -> V:
        self._limit = n
        return self

    def _build_aggregate_pipeline(self: V) -> list[dict[str, Any]]:
        lookups = super()._build_aggregate_pipeline()
        result: list[dict[str, Any]] = []
        if self._match is not None:
            result.append({'$match': self._match})
        if self._sort is not None:
            result.append({'$sort': self._sort})
        if self._limit is not None:
            result.append({'$limit': self._limit})
        if self._skip is not None:
            result.append({'$skip': self._skip})
        result.extend(lookups)
        return result

    def _exec(self: V) -> list[T]:
        pipeline = self._build_aggregate_pipeline()
        collection = Connection.get_collection(self._cls)
        cursor = collection.aggregate(pipeline)
        results = [result for result in cursor]
        return Decoder().decode_root_list(results, self._cls)


class ListQuery(BaseListQuery[T]):
    """Query a list of objects.
    """

    def exec(self) -> list[T]:
        return self._exec()

    def __await__(self) -> Generator[None, None, list[T]]:
        yield
        return self.exec()


class SingleQuery(BaseListQuery[T]):
    """Queries only one object from the query.
    """

    def exec(self) -> T:
        self._limit = 1
        results = self._exec()
        if len(results) == 0:
            raise ObjectNotFoundException('object is not found')
        return results[0]

    def __await__(self) -> Generator[None, None, T]:
        yield
        return self.exec()

    @property
    def optional(self) -> OptionalSingleQuery:
        query = OptionalSingleQuery(cls=self._cls)
        query.subqueries = self.subqueries
        query._match = self._match
        query._sort = self._sort
        query._skip = self._skip
        query._limit = self._limit
        query._use_pick = self._use_pick
        query._pick = self._pick
        query._use_omit = self._use_omit
        query._omit = self._omit
        return query


class OptionalSingleQuery(BaseListQuery[T]):
    """Queries only one object from the query. Returns None if not found.
    """

    def exec(self) -> Optional[T]:
        self._limit = 1
        results = self._exec()
        if len(results) == 0:
            return None
        return results[0]

    def __await__(self) -> Generator[None, None, Optional[T]]:
        yield
        return self.exec()


class BaseIDQuery(BaseQuery[T]):
    """Query collection from object id.
    """

    def __init__(self: U, cls: type[T], id: Union[str, ObjectId]) -> None:
        super().__init__(cls)
        self._id = id

    def _build_aggregate_pipeline(self: BaseIDQuery) -> list[dict[str, Any]]:
        lookups = super()._build_aggregate_pipeline()
        result: list[dict[str, Any]] = []
        result.append({'$match': {'_id': ObjectId(self._id)}})
        result.extend(lookups)
        return result

    def _exec(self) -> Optional[T]:
        pipeline = self._build_aggregate_pipeline()
        collection = Connection.get_collection(self._cls)
        cursor = collection.aggregate(pipeline)
        results = [result for result in cursor]
        if len(results) == 0:
            return None
        return Decoder().decode_root(results[0], self._cls)


class IDQuery(BaseIDQuery[T]):
    """Query collection from object id. Raises ObjectNotFoundException if not
    found.
    """

    def exec(self) -> T:
        result = self._exec()
        if result is not None:
            return result
        raise ObjectNotFoundException(
            f'{self._cls.__name__}(_id={self._id}) not found.')

    def __await__(self) -> Generator[None, None, T]:
        yield
        return self.exec()

    @property
    def optional(self) -> OptionalIDQuery:
        new_query = OptionalIDQuery(cls=self._cls, id=self._id)
        new_query.subqueries = self.subqueries
        return new_query


class OptionalIDQuery(BaseIDQuery[T]):
    """Query collection from object id. Returns None if not found.
    """

    def exec(self) -> Optional[T]:
        return self._exec()

    def __await__(self) -> Generator[None, None, Optional[T]]:
        yield
        return self.exec()


class ExistQuery(BaseListQuery[T]):

    def exec(self) -> bool:
        collection = Connection.get_collection(self._cls)
        result = collection.count_documents(self._match or {}, limit=1)
        return False if result == 0 else True

    def __await__(self) -> Generator[None, None, bool]:
        yield
        return self.exec()


class QueryIterator(Generic[T]):

    def __init__(self, cls: type[T], cursor: Cursor):
        self.cls = cls
        self.cursor = cursor
        self.graph = MarkGraph()

    def __iter__(self):
        return self

    def __next__(self) -> T:
        value = self.cursor.__next__()
        return Decoder().decode_root(value, self.cls, self.graph)


class IterateQuery(BaseListQuery[T]):

    def exec(self) -> Iterator[T]:
        pipeline = self._build_aggregate_pipeline()
        collection = Connection.get_collection(self._cls)
        cursor = collection.aggregate(pipeline)
        return QueryIterator(cls=self._cls, cursor=cursor)

    def __await__(self) -> Generator[None, None, Iterator[T]]:
        yield
        return self.exec()


# class CountQuery(BaseListQuery[T]):

#     def exec(self) -> int:
#         pass

#     def __await__(self) -> Generator[None, None, int]:
#         yield
#         return self.exec()



from __future__ import annotations
from typing import Optional
from unittest import TestCase
from jsonclasses.exceptions import UniqueConstraintException
from jsonclasses_pymongo import Connection
from tests.classes.simple_person import SimplePerson
from tests.classes.simple_singer import SimpleSinger


class TestUniqueRecord(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        connection = Connection('simple')
        connection.set_url('mongodb://localhost:27017/simple')
        connection.connect()

    @classmethod
    def tearDownClass(cls) -> None:
        connection = Connection('simple')
        connection.disconnect()

    def setUp(self) -> None:
        collection = Connection.get_collection(SimplePerson)
        collection.delete_many({})
        collection = Connection.get_collection(SimpleSinger)
        collection.delete_many({})

    def test_save_raises_if_violate_unique_rule(self):
        one = SimplePerson(name='Tsuan Tsiu')
        one.save()
        two = SimplePerson(name='Tsuan Tsiu')
        self.assertRaisesRegex(
            UniqueConstraintException,
            "Value 'Tsuan Tsiu' at 'name' is not unique.",
            two.save)

    def test_save_wont_raise_if_value_is_optional_and_is_null(self):
        one = SimpleSinger(name=None)
        one.save()
        two = SimpleSinger(name=None)
        two.save()
        one = SimplePerson(name='Tsuan Tsiu')
        one.save()
        two = SimplePerson(name='Tsuan Tsiu')
        self.assertRaisesRegex(
            UniqueConstraintException,
            "Value 'Tsuan Tsiu' at 'name' is not unique.",
            two.save)



from __future__ import annotations
from datetime import datetime
from unittest import TestCase
from bson.objectid import ObjectId
from jsonclasses_pymongo import Connection
from tests.classes.simple_song import SimpleSong
from tests.classes.simple_artist import SimpleArtist
from tests.classes.linked_author import LinkedAuthor
from tests.classes.linked_post import LinkedPost
from tests.classes.linked_profile_user import LinkedProfile, LinkedUser
from tests.classes.linked_favorite import LinkedCourse, LinkedStudent


class TestSave(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        connection = Connection('simple')
        connection.set_url('mongodb://localhost:27017/simple')
        connection.connect()
        connection = Connection('linked')
        connection.set_url('mongodb://localhost:27017/linked')
        connection.connect()

    @classmethod
    def tearDownClass(cls) -> None:
        connection = Connection('simple')
        connection.disconnect()
        connection = Connection('linked')
        connection.disconnect()

    def setUp(self) -> None:
        collection = Connection.get_collection(SimpleSong)
        collection.delete_many({})
        collection = Connection.get_collection(SimpleArtist)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedAuthor)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedPost)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedProfile)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedUser)
        collection.delete_many({})
        collection.delete_many({})
        collection = Connection.get_collection(LinkedStudent)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedCourse)
        collection.delete_many({})
        collection = Connection('linked').collection('linkedcoursesstudents'
                                                     'linkedstudentscourses')
        collection.delete_many({})

    def test_object_is_saved_into_database(self):
        song = SimpleSong(name='Long', year=2020, artist='Thao')
        song.save()
        collection = Connection.get_collection(SimpleSong)
        self.assertEqual(collection.count_documents({}), 1)
        for item in collection.find():
            self.assertEqual(item['name'], 'Long')
            self.assertEqual(item['year'], 2020)
            self.assertEqual(item['artist'], 'Thao')
            self.assertIsInstance(item['_id'], ObjectId)
            self.assertIsInstance(item['updatedAt'], datetime)
            self.assertIsInstance(item['createdAt'], datetime)
            self.assertEqual(set(item.keys()),
                             {'_id', 'name', 'year', 'artist', 'updatedAt',
                              'createdAt', 'deletedAt'})

    def test_object_with_enum_is_saved_into_database(self):
        artist = SimpleArtist(name='Kaosai', gender='MALE')
        artist.save()
        collection = Connection.get_collection(SimpleArtist)
        self.assertEqual(collection.count_documents({}), 1)
        for item in collection.find():
            self.assertEqual(item['name'], 'Kaosai')
            self.assertEqual(item['gender'], 1)
            self.assertIsInstance(item['_id'], ObjectId)
            self.assertIsInstance(item['updatedAt'], datetime)
            self.assertIsInstance(item['createdAt'], datetime)
            self.assertEqual(set(item.keys()),
                             {'_id', 'name', 'gender', 'updatedAt',
                              'createdAt', 'deletedAt'})

    def test_linked_objects_are_saved_at_the_same_time(self):
        input = {
            'name': 'Ti',
            'posts': [
                {
                    'title': 'Bo Lo Iong',
                    'content': 'Pieng Iu'
                },
                {
                    'title': 'Bo Lo Iong',
                    'content': 'Pieng Iu'
                }
            ]
        }
        author = LinkedAuthor(**input)
        author.save()
        collection = Connection.get_collection(LinkedAuthor)
        self.assertEqual(collection.count_documents({}), 1)
        oid = ObjectId()
        for item in collection.find():
            self.assertEqual(item['name'], 'Ti')
            self.assertIsInstance(item['_id'], ObjectId)
            oid = item['_id']
            self.assertIsInstance(item['updatedAt'], datetime)
            self.assertIsInstance(item['createdAt'], datetime)
            self.assertEqual(set(item.keys()),
                             {'_id', 'name', 'updatedAt',
                              'createdAt', 'deletedAt'})
        collection = Connection.get_collection(LinkedPost)
        self.assertEqual(collection.count_documents({}), 2)
        for item in collection.find():
            self.assertEqual(item['title'], 'Bo Lo Iong')
            self.assertEqual(item['content'], 'Pieng Iu')
            self.assertEqual(item['authorId'], oid)
            self.assertIsInstance(item['_id'], ObjectId)
            self.assertIsInstance(item['updatedAt'], datetime)
            self.assertIsInstance(item['createdAt'], datetime)
            self.assertEqual(set(item.keys()),
                             {'_id', 'title', 'content', 'authorId',
                              'updatedAt', 'createdAt', 'deletedAt'})

    def test_object_is_not_new_after_saved(self):
        song = SimpleSong(name='Long', year=2020, artist='Thao')
        self.assertEqual(song.is_new, True)
        self.assertEqual(song.is_modified, False)
        self.assertEqual(song.modified_fields, ())
        song.save()
        self.assertEqual(song.is_new, False)
        self.assertEqual(song.is_modified, False)
        self.assertEqual(song.modified_fields, ())
        song.name = 'Lao'
        self.assertEqual(song.is_new, False)
        self.assertEqual(song.is_modified, True)
        self.assertEqual(song.modified_fields, ('name',))

    def test_update_can_update_database(self):
        song = SimpleSong(name='Tsong Khang', year=2020, artist='Tsao')
        song.save()
        collection = Connection.get_collection(SimpleSong)
        self.assertEqual(collection.count_documents({}), 1)
        song.name = 'Lao'
        song.save()
        self.assertEqual(collection.count_documents({}), 1)
        for item in collection.find():
            self.assertEqual(item['name'], 'Lao')
            self.assertEqual(item['year'], 2020)
            self.assertEqual(item['artist'], 'Tsao')
            self.assertIsInstance(item['_id'], ObjectId)
            self.assertIsInstance(item['updatedAt'], datetime)
            self.assertIsInstance(item['createdAt'], datetime)
            self.assertEqual(set(item.keys()),
                             {'_id', 'name', 'year', 'artist', 'updatedAt',
                              'createdAt', 'deletedAt'})

    def test_linked_object_is_not_new_after_saved(self):
        input = {
            'name': 'Ti',
            'posts': [
                {
                    'title': 'Bo Lo Iong',
                    'content': 'Pieng Iu'
                },
                {
                    'title': 'Bo Lo Iong',
                    'content': 'Pieng Iu'
                }
            ]
        }
        author = LinkedAuthor(**input)
        self.assertEqual(author.is_new, True)
        self.assertEqual(author.posts[0].is_new, True)
        self.assertEqual(author.posts[1].is_new, True)
        author.save()
        self.assertEqual(author.is_new, False)
        self.assertEqual(author.posts[0].is_new, False)
        self.assertEqual(author.posts[1].is_new, False)

    def test_set_local_key_to_none_is_saved(self):
        post = LinkedPost(title='P1', content='P2')
        author = LinkedAuthor(name='A1')
        post.author = author
        post.save()
        post.author_id = None
        post.save()
        collection = Connection.get_collection(LinkedPost)
        for item in collection.find({}):
            self.assertEqual(item['authorId'], None)

    def test_set_local_key_to_id_is_saved(self):
        post = LinkedPost(title='P1', content='P2')
        author = LinkedAuthor(name='A1')
        post.save()
        author.save()
        post.author_id = author.id
        post.save()
        collection = Connection.get_collection(LinkedPost)
        for item in collection.find({}):
            self.assertEqual(item['authorId'], ObjectId(author.id))

    def test_alter_local_key_is_saved(self):
        post = LinkedPost(title='P1', content='P2')
        a1 = LinkedAuthor(name='A1')
        post.author = a1
        post.save()
        a2 = LinkedAuthor(name='A2')
        post.author_id = a2.id
        post.save()
        collection = Connection.get_collection(LinkedPost)
        for item in collection.find({}):
            self.assertEqual(item['authorId'], ObjectId(a2.id))

    def test_local_key_from_param_is_saved(self):
        a1 = LinkedAuthor(name='A1')
        a1.save()
        post = LinkedPost(title='P1', content='P2', authorId=a1.id)
        post.save()
        collection = Connection.get_collection(LinkedPost)
        for item in collection.find({}):
            self.assertEqual(item['authorId'], ObjectId(a1.id))

    def test_local_key_from_set_is_altered_and_saved(self):
        a1 = LinkedAuthor(name='A1')
        a1.save()
        post = LinkedPost(title='P1', content='P2', authorId=a1.id)
        post.save()
        collection = Connection.get_collection(LinkedPost)
        for item in collection.find({}):
            self.assertEqual(item['authorId'], ObjectId(a1.id))
        a2 = LinkedAuthor(name='A2')
        post.set(authorId=a2.id)
        post.save()
        for item in collection.find({}):
            self.assertEqual(item['authorId'], ObjectId(a2.id))

    def test_1l_1f_object_unlink_is_saved(self):
        profile = LinkedProfile(name='p')
        user = LinkedUser(name='u')
        profile.user = user
        profile.save()
        profile.user = None
        profile.save()
        self.assertEqual(profile.is_new, False)
        self.assertEqual(profile.is_modified, False)
        self.assertEqual(profile.modified_fields, ())
        self.assertEqual(user.is_new, False)
        self.assertEqual(user.is_modified, False)
        self.assertEqual(user.modified_fields, ())
        collection = Connection.get_collection(LinkedProfile)
        for item in collection.find({}):
            self.assertEqual(item['userId'], None)

    def test_1f_1l_object_unlink_is_saved(self):
        user = LinkedUser(name='u')
        profile = LinkedProfile(name='p')
        user.profile = profile
        user.save()
        user.profile = None
        user.save()
        self.assertEqual(profile.is_new, False)
        self.assertEqual(profile.is_modified, False)
        self.assertEqual(profile.modified_fields, ())
        self.assertEqual(user.is_new, False)
        self.assertEqual(user.is_modified, False)
        self.assertEqual(user.modified_fields, ())
        collection = Connection.get_collection(LinkedProfile)
        for item in collection.find({}):
            self.assertEqual(item['userId'], None)

    def test_1_many_unlink_is_saved(self):
        author = LinkedAuthor(name='A')
        post = LinkedPost(title='T', content='C')
        author.posts.append(post)
        author.save()
        author.posts.remove(post)
        author.save()
        self.assertEqual(author.is_new, False)
        self.assertEqual(author.is_modified, False)
        self.assertEqual(author.modified_fields, ())
        self.assertEqual(post.is_new, False)
        self.assertEqual(post.is_modified, False)
        self.assertEqual(post.modified_fields, ())
        collection = Connection.get_collection(LinkedPost)
        for item in collection.find({}):
            self.assertEqual(item['authorId'], None)

    def test_many_1_unlink_is_saved(self):
        author = LinkedAuthor(name='A')
        post = LinkedPost(title='T', content='C')
        post.author = author
        post.save()
        post.author_id = None
        post.save()
        self.assertEqual(author.is_new, False)
        self.assertEqual(author.is_modified, False)
        self.assertEqual(author.modified_fields, ())
        self.assertEqual(post.is_new, False)
        self.assertEqual(post.is_modified, False)
        self.assertEqual(post.modified_fields, ())
        collection = Connection.get_collection(LinkedPost)
        for item in collection.find({}):
            self.assertEqual(item['authorId'], None)

    def test_many_many_unlink_is_saved(self):
        course1 = LinkedCourse(name='C1')
        course2 = LinkedCourse(name='C2')
        student1 = LinkedStudent(name='S1')
        student2 = LinkedStudent(name='S2')
        course1.students = [student1, student2]
        course2.students = [student1, student2]
        course1.save()
        course1.students.remove(student1)
        course1.save()
        collname = 'linkedcoursesstudentslinkedstudentscourses'
        collection = Connection('linked').collection(collname)
        self.assertEqual(collection.count_documents({}), 3)



from __future__ import annotations
from typing import TYPE_CHECKING
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import pymongo, BaseObject
if TYPE_CHECKING:
    from .linked_post import LinkedPost


@pymongo
@jsonclass(class_graph='linked')
class LinkedAuthor(BaseObject):
    name: str
    posts: list[LinkedPost] = types.nonnull.listof('LinkedPost') \
                                   .linkedby('author')



from __future__ import annotations
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import pymongo, BaseObject


@pymongo
@jsonclass(class_graph='linked')
class LinkedOrder(BaseObject):
    name: str
    user: LinkedBuyer = types.instanceof('LinkedBuyer').linkto


@pymongo
@jsonclass(class_graph='linked')
class LinkedBuyer(BaseObject):
    name: str
    orders: list[LinkedOrder] = types.listof('LinkedOrder').linkedby('user') \
                                     .cascade


@pymongo
@jsonclass(class_graph='linked')
class LinkedCOrder(BaseObject):
    name: str
    user: LinkedCBuyer = types.instanceof('LinkedCBuyer').linkto.cascade


@pymongo
@jsonclass(class_graph='linked')
class LinkedCBuyer(BaseObject):
    name: str
    orders: list[LinkedCOrder] = types.listof('LinkedCOrder').linkedby('user')\
                                      .cascade



from __future__ import annotations
from unittest import TestCase
from typing import List, Dict
from datetime import date, datetime
from bson import ObjectId
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import BaseObject, pymongo
from jsonclasses_pymongo.decoder import Decoder


class TestDecoder(TestCase):

    def test_decode_str_into_str(self):
        @pymongo
        @jsonclass
        class SimpleDecodeStr(BaseObject):
            val1: str
            val2: str
        data = {
            '_id': ObjectId(),
            'createdAt': datetime.now(),
            'updatedAt': datetime.now(),
            'val1': '12345',
            'val2': '67890'
        }
        instance = Decoder().decode_root(data, SimpleDecodeStr)
        self.assertEqual(instance.id, str(data['_id']))
        self.assertEqual(instance.val1, '12345')
        self.assertEqual(instance.val2, '67890')
        self.assertIsInstance(instance.val1, str)
        self.assertIsInstance(instance.val2, str)

    def test_decode_int_into_int(self):
        @pymongo
        @jsonclass
        class SimpleDecodeInt(BaseObject):
            val1: int
            val2: int
        data = {
            '_id': ObjectId(),
            'createdAt': datetime.now(),
            'updatedAt': datetime.now(),
            'val1': 12345,
            'val2': 67890
        }
        instance = Decoder().decode_root(data, SimpleDecodeInt)
        self.assertEqual(instance.id, str(data['_id']))
        self.assertEqual(instance.val1, 12345)
        self.assertEqual(instance.val2, 67890)

    def test_decode_float_into_float(self):
        @pymongo
        @jsonclass
        class SimpleDecodeFloat(BaseObject):
            val1: float
            val2: float
        data = {
            '_id': ObjectId(),
            'createdAt': datetime.now(),
            'updatedAt': datetime.now(),
            'val1': 12345.6,
            'val2': 67890.1
        }
        instance = Decoder().decode_root(data, SimpleDecodeFloat)
        self.assertEqual(instance.id, str(data['_id']))
        self.assertEqual(instance.val1, 12345.6)
        self.assertEqual(instance.val2, 67890.1)

    def test_decode_bool_into_bool(self):
        @pymongo
        @jsonclass
        class SimpleDecodeBool(BaseObject):
            val1: bool
            val2: bool
        data = {
            '_id': ObjectId(),
            'createdAt': datetime.now(),
            'updatedAt': datetime.now(),
            'val1': True,
            'val2': False
        }
        instance = Decoder().decode_root(data, SimpleDecodeBool)
        self.assertEqual(instance.id, str(data['_id']))
        self.assertEqual(instance.val1, True)
        self.assertEqual(instance.val2, False)

    def test_decode_datetime_into_date(self):
        @pymongo
        @jsonclass
        class SimpleDecodeDate(BaseObject):
            val1: date
            val2: date
        data = {
            '_id': ObjectId(),
            'createdAt': datetime.now(),
            'updatedAt': datetime.now(),
            'val1': datetime(2012, 9, 5, 0, 0, 0),
            'val2': datetime(2020, 9, 5, 0, 0, 0),
        }
        instance = Decoder().decode_root(data, SimpleDecodeDate)
        self.assertEqual(instance.id, str(data['_id']))
        self.assertEqual(instance.val1, date(2012, 9, 5))
        self.assertEqual(instance.val2, date(2020, 9, 5))

    def test_decode_datetime_into_datetime(self):
        @pymongo
        @jsonclass
        class SimpleDecodeDatetime(BaseObject):
            val1: datetime
            val2: datetime
        data = {
            '_id': ObjectId(),
            'createdAt': datetime.now(),
            'updatedAt': datetime.now(),
            'val1': datetime(2012, 9, 5, 6, 25, 0),
            'val2': datetime(2020, 9, 5, 8, 25, 0),
        }
        instance = Decoder().decode_root(data, SimpleDecodeDatetime)
        self.assertEqual(instance.id, str(data['_id']))
        self.assertEqual(instance.val1, datetime(2012, 9, 5, 6, 25, 0))
        self.assertEqual(instance.val2, datetime(2020, 9, 5, 8, 25, 0))

    def test_decode_embedded_list(self):
        @pymongo
        @jsonclass
        class SimpleDecodeList(BaseObject):
            vals: List[int]
        data = {
            '_id': ObjectId(),
            'createdAt': datetime.now(),
            'updatedAt': datetime.now(),
            'vals': [1, 2, 3, 4, 5]
        }
        instance = Decoder().decode_root(data, SimpleDecodeList)
        self.assertEqual(instance.id, str(data['_id']))
        self.assertEqual(instance.vals, [1, 2, 3, 4, 5])

    def test_decode_local_keys_list(self):
        @pymongo
        @jsonclass
        class SimpleDecodeLocalKeyListAddress(BaseObject):
            city: str
            owner: SimpleDecodeLocalKeyList = types.instanceof(
                'SimpleDecodeLocalKeyList').linkedby('address')

        @pymongo
        @jsonclass
        class SimpleDecodeLocalKeyList(BaseObject):
            addresses: List[SimpleDecodeLocalKeyListAddress] = (types.linkto
                .listof(SimpleDecodeLocalKeyListAddress))
        data = {
            '_id': ObjectId(),
            'createdAt': datetime.now(),
            'updatedAt': datetime.now(),
            'addressesIds': [ObjectId(), ObjectId()]
        }
        instance = Decoder().decode_root(data, SimpleDecodeLocalKeyList)
        self.assertEqual(instance.id, str(data['_id']))
        self.assertEqual(getattr(instance, 'addresses_ids')
                         [0], str(data['addressesIds'][0]))
        self.assertEqual(getattr(instance, 'addresses_ids')
                         [1], str(data['addressesIds'][1]))

    def test_decode_embedded_dict(self):
        @pymongo
        @jsonclass
        class SimpleDecodeDict(BaseObject):
            vals: Dict[str, int]
        data = {
            '_id': ObjectId(),
            'createdAt': datetime.now(),
            'updatedAt': datetime.now(),
            'vals': {'one': 1, 'two': 2}
        }
        instance = Decoder().decode_root(data, SimpleDecodeDict)
        self.assertEqual(instance.id, str(data['_id']))
        self.assertEqual(instance.vals, {'one': 1, 'two': 2})

    def test_decode_embedded_shape(self):
        @pymongo
        @jsonclass
        class SimpleDecodeShape(BaseObject):
            vals: Dict[str, int] = types.shape({
                'one': types.int,
                'two': types.int
            })
        data = {
            '_id': ObjectId(),
            'createdAt': datetime.now(),
            'updatedAt': datetime.now(),
            'vals': {'one': 1, 'two': 2}
        }
        instance = Decoder().decode_root(data, SimpleDecodeShape)
        self.assertEqual(instance.id, str(data['_id']))
        self.assertEqual(instance.vals, {'one': 1, 'two': 2})

    def test_decode_embedded_instance(self):
        @pymongo
        @jsonclass
        class SimpleDecodeInstanceAddress(BaseObject):
            city: str

        @pymongo
        @jsonclass
        class SimpleDecodeInstance(BaseObject):
            address: SimpleDecodeInstanceAddress
        data = {
            '_id': ObjectId(),
            'createdAt': datetime.now(),
            'updatedAt': datetime.now(),
            'address': {
                '_id': ObjectId(),
                'createdAt': datetime.now(),
                'updatedAt': datetime.now(),
                'city': 'Shanghai'
            }
        }
        instance = Decoder().decode_root(data, SimpleDecodeInstance)
        self.assertEqual(instance.id, str(data['_id']))
        self.assertEqual(instance.address.id, str(data['address']['_id']))
        self.assertEqual(instance.address.city, "Shanghai")

    def test_decode_local_key_instance(self):
        @pymongo
        @jsonclass
        class SimpleDecodeLocalKeyInstanceAddress(BaseObject):
            city: str
            owner: SimpleDecodeLocalKeyInstance = types.instanceof(
                'SimpleDecodeLocalKeyInstance').linkedby('address')

        @pymongo
        @jsonclass
        class SimpleDecodeLocalKeyInstance(BaseObject):
            address: SimpleDecodeLocalKeyInstanceAddress = (types
                    .linkto.instanceof(
                        SimpleDecodeLocalKeyInstanceAddress))
        data = {
            '_id': ObjectId(),
            'createdAt': datetime.now(),
            'updatedAt': datetime.now(),
            'addressId': ObjectId()
        }
        instance = Decoder().decode_root(data, SimpleDecodeLocalKeyInstance)
        self.assertEqual(instance.id, str(data['_id']))
        self.assertEqual(getattr(instance, 'address_id'),
                         str(data['addressId']))

    def test_decode_camelized_dict_keys(self):
        @pymongo
        @jsonclass
        class MediumDecodeCamelizeDictKeys(BaseObject):
            val: Dict[str, str] = types.dictof(types.str)
        data = {
            'val': {
                'keyOne': 'val_one',
                'keyTwo': 'val_two'
            }
        }
        instance = Decoder().decode_root(data, MediumDecodeCamelizeDictKeys)
        self.assertEqual(
            instance.val, {'key_one': 'val_one', 'key_two': 'val_two'})

    def test_decode_uncamelized_dict_keys(self):
        @pymongo(camelize_db_keys=False)
        @jsonclass
        class MediumDecodeUncamelizeDictKeys(BaseObject):
            val: Dict[str, str] = types.dictof(types.str)
        data = {
            'val': {
                'keyOne': 'val_one',
                'keyTwo': 'val_two'
            }
        }
        instance = Decoder().decode_root(data, MediumDecodeUncamelizeDictKeys)
        self.assertEqual(
            instance.val, {'keyOne': 'val_one', 'keyTwo': 'val_two'})

    def test_decode_camelized_shape_keys(self):
        @pymongo
        @jsonclass
        class MediumDecodeCamelizeShapeKeys(BaseObject):
            val: Dict[str, str] = types.shape({
                'key_one': types.str,
                'key_two': types.str
            })
        data = {
            'val': {
                'keyOne': 'val_one',
                'keyTwo': 'val_two'
            }
        }
        instance = Decoder().decode_root(data, MediumDecodeCamelizeShapeKeys)
        self.assertEqual(
            instance.val, {'key_one': 'val_one', 'key_two': 'val_two'})

    def test_decode_uncamelized_shape_keys(self):
        @pymongo(camelize_db_keys=False)
        @jsonclass
        class MediumDecodeUncamelizeShapeKeys(BaseObject):
            val: Dict[str, str] = types.shape({
                'key_one': types.str,
                'key_two': types.str
            })
        data = {
            'val': {
                'key_one': 'val_one',
                'key_two': 'val_two'
            }
        }
        instance = Decoder().decode_root(data, MediumDecodeUncamelizeShapeKeys)
        self.assertEqual(
            instance.val, {'key_one': 'val_one', 'key_two': 'val_two'})



from __future__ import annotations
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import pymongo, BaseObject


@pymongo
@jsonclass(class_graph='linked')
class LinkedAccount(BaseObject):
    name: str
    balance: LinkedBalance = types.instanceof('LinkedBalance').linkto.cascade


@pymongo
@jsonclass(class_graph='linked')
class LinkedBalance(BaseObject):
    name: str
    account: LinkedAccount = types.instanceof('LinkedAccount') \
                                  .linkedby('balance').cascade



from __future__ import annotations
from datetime import datetime
from unittest import TestCase
from bson.objectid import ObjectId
from jsonclasses_pymongo import Connection
from tests.classes.simple_song import SimpleSong
from tests.classes.simple_artist import SimpleArtist
from tests.classes.linked_author import LinkedAuthor
from tests.classes.linked_post import LinkedPost


class TestQuery(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        connection = Connection('simple')
        connection.set_url('mongodb://localhost:27017/simple')
        connection.connect()
        connection = Connection('linked')
        connection.set_url('mongodb://localhost:27017/linked')
        connection.connect()

    @classmethod
    def tearDownClass(cls) -> None:
        connection = Connection('simple')
        connection.disconnect()
        connection = Connection('linked')
        connection.disconnect()

    def setUp(self) -> None:
        collection = Connection.get_collection(SimpleSong)
        collection.delete_many({})
        collection = Connection.get_collection(SimpleArtist)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedAuthor)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedPost)
        collection.delete_many({})

    def test_query_objects_from_database(self):
        song0 = SimpleSong(name='Long', year=2020, artist='Thao')
        song0.save()
        song1 = SimpleSong(name='Long', year=2020, artist='Thao')
        song1.save()
        songs = SimpleSong.find().exec()
        self.assertEqual(song0.name, songs[0].name)
        self.assertEqual(song0.year, songs[0].year)
        self.assertEqual(song0.artist, songs[0].artist)
        self.assertEqual(song0.id, songs[0].id)
        self.assertGreaterEqual(song0.created_at, songs[0].created_at)
        self.assertGreaterEqual(song0.updated_at, songs[0].updated_at)
        self.assertEqual(song0.deleted_at, songs[0].deleted_at)
        self.assertEqual(song1.name, songs[1].name)
        self.assertEqual(song1.year, songs[1].year)
        self.assertEqual(song1.artist, songs[1].artist)
        self.assertEqual(song1.id, songs[1].id)
        self.assertGreaterEqual(song1.created_at, songs[1].created_at)
        self.assertGreaterEqual(song1.updated_at, songs[1].updated_at)
        self.assertEqual(song1.deleted_at, songs[1].deleted_at)

    def test_query_object_from_database(self):
        song0 = SimpleSong(name='Long', year=2020, artist='Thao')
        song0.save()
        song1 = SimpleSong(name='Long', year=2020, artist='Thao')
        song1.save()
        song = SimpleSong.one().exec()
        self.assertEqual(song0.name, song.name)
        self.assertEqual(song0.year, song.year)
        self.assertEqual(song0.artist, song.artist)
        self.assertEqual(song0.id, song.id)
        self.assertGreaterEqual(song0.created_at, song.created_at)
        self.assertGreaterEqual(song0.updated_at, song.updated_at)
        self.assertEqual(song0.deleted_at, song.deleted_at)

    def test_query_object_with_id(self):
        song = SimpleSong(name='Long', year=2020, artist='Thao')
        song.save()
        result = SimpleSong.id(str(song.id)).exec()
        self.assertEqual(song.name, result.name)
        self.assertEqual(song.year, result.year)
        self.assertEqual(song.artist, result.artist)
        self.assertEqual(song.id, result.id)
        self.assertGreaterEqual(song.created_at, result.created_at)
        self.assertGreaterEqual(song.updated_at, result.updated_at)
        self.assertEqual(song.deleted_at, result.deleted_at)



from __future__ import annotations
from typing import Callable, ClassVar, Optional, TYPE_CHECKING, TypeVar
from os import getcwd, path
from pymongo.mongo_client import MongoClient
from pymongo.database import Database
from pymongo.collection import Collection
from inflection import parameterize, camelize
if TYPE_CHECKING:
    from .pymongo_object import PymongoObject
    T = TypeVar('T', bound=PymongoObject, covariant=False)


ConnectedCallback = Callable[[Collection], None]


class Connection:
    _graph_map: dict[str, Connection] = {}
    _initialized_map: dict[str, bool] = {}

    def __new__(cls: type[Connection], graph_name: str) -> Connection:
        if not cls._graph_map.get(graph_name):
            cls._graph_map[graph_name] = super(Connection, cls).__new__(cls)
        return cls._graph_map[graph_name]

    def __init__(self: Connection, graph_name: str) -> None:
        if self.__class__._initialized_map.get(graph_name):
            return
        self._graph_name: str = graph_name
        self._url: Optional[str] = None
        self._client: Optional[MongoClient] = None
        self._database: Optional[Database] = None
        self._collections: dict[str, Collection] = {}
        self._connection_callbacks: dict[str, ConnectedCallback] = {}
        self._connected: bool = False
        self.__class__._initialized_map[graph_name] = True
        return None

    @property
    def graph_name(self: Connection) -> str:
        return self._graph_name

    @property
    def url(self: Connection) -> str:
        if self._url:
            return self._url
        return self._generate_default_url()

    def set_url(self: Connection, url: str) -> None:
        self._url = url

    def _generate_default_url(self: Connection) -> str:
        base = 'mongodb://localhost:27017/'
        cwd = getcwd()
        proj = camelize(parameterize(path.basename(cwd))).lower()
        self._url = base + proj
        return self._url

    @property
    def client(self: Connection) -> MongoClient:
        if self._client is not None:
            return self._client
        self.connect()
        return self._client

    @property
    def database(self: Connection) -> Database:
        if self._database is not None:
            return self._database
        self.connect()
        return self._database

    def connect(self: Connection) -> None:
        self._client = MongoClient(self.url)
        self._database = self._client.get_database()
        self._connected = True
        for name, callback in self._connection_callbacks.items():
            self._call_callback(name, callback)

    def disconnect(self: Connection) -> None:
        if self._client is not None:
            self._client.close()
            self._client = None
            self._database = None
            self._collections = {}
            self._connected = False

    @property
    def connected(self: Collection) -> bool:
        return self._connected

    def collection(self: Connection, name: str) -> Collection:
        if self._collections.get(name):
            return self._collections[name]
        coll = self.database.get_collection(name)
        self._collections[name] = coll
        return coll

    def add_connected_callback(self: Connection,
                               name: str,
                               callback: ConnectedCallback) -> None:
        self._connection_callbacks[name] = callback
        if self._client:
            self._call_callback(name, callback)

    def _call_callback(self: Connection,
                       name: str,
                       callback: ConnectedCallback) -> None:
        callback(self.collection(name))

    def collection_from(self: Connection,
                        cls: type[T]) -> Collection:
        coll_name = cls.dbconf.collection_name
        return self.collection(coll_name)

    default: ClassVar[Connection]

    @classmethod
    def get_collection(cls: type[Connection],
                       pmcls: type[T]) -> Collection:
        graph = pmcls.definition.config.class_graph.name
        connection = Connection(graph)
        return connection.collection_from(pmcls)

    @classmethod
    def from_class(cls: type[Connection],
                   pmcls: type[T]) -> Connection:
        return Connection(pmcls.definition.config.class_graph.name)


Connection.default = Connection('default')



"""
This module contains `jsonclass`, the decorator for JSON Classes.
"""
from typing import Optional, Union, Callable, overload, cast
from jsonclasses.jsonclass_object import JSONClassObject
from .pymongo_object import PymongoObject
from .pymongofy import pymongofy
from .dbconf import DBConf


@overload
def pymongo(cls: type) -> type: ...


@overload
def pymongo(
    cls: None,
    collection_name: Optional[str] = None,
    camelize_db_keys: Optional[bool] = None,
) -> Callable[[type], type[PymongoObject]]: ...


@overload
def pymongo(
    cls: type,
    collection_name: Optional[str] = None,
    camelize_db_keys: Optional[bool] = None,
) -> type[PymongoObject]: ...


def pymongo(
    cls: Optional[type[JSONClassObject]] = None,
    collection_name: Optional[str] = None,
    camelize_db_keys: Optional[bool] = None,
) -> Union[Callable[[type], type[PymongoObject]], type[PymongoObject]]:
    """The pymongo object class decorator. To declare a jsonclass class, use
    this syntax:

        @pymongo
        @jsonclass
        class MyObject:
            my_field_one: str
            my_field_two: bool
    """
    if cls is not None:
        if not isinstance(cls, type):
            raise ValueError('@pymongo should be used to decorate a class.')
        conf = DBConf(cls, cls.definition.config,
                      collection_name, camelize_db_keys)
        cls.dbconf = conf
        return cast(type[PymongoObject], pymongofy(cls))
    else:
        def parametered_jsonclass(cls):
            return pymongo(
                cls,
                collection_name=collection_name,
                camelize_db_keys=camelize_db_keys)
        return parametered_jsonclass



"""This module defines `PymongoObject`, the protocol that @pymongo decorated
jsonclass object should confirm to.
"""
from __future__ import annotations
from typing import TypeVar, ClassVar, Any, Union, TYPE_CHECKING
from bson.objectid import ObjectId
from jsonclasses.jsonclass_object import JSONClassObject
if TYPE_CHECKING:
    from .dbconf import DBConf
    from .query import (BaseQuery, ListQuery, IDQuery, SingleQuery, ExistQuery,
                        IterateQuery)


T = TypeVar('T', bound='PymongoObject')


class PymongoObject(JSONClassObject):
    """The `PymongoObject` protocol defines methods that @pymongo decorated
    jsonclass object should confirm to.
    """

    dbconf: ClassVar[DBConf]
    """The configuration user passed to JSON class through the jsonclass
    decorator.
    """

    @classmethod
    def find(cls: type[T], **kwargs: Any) -> ListQuery[T]:
        ...

    @classmethod
    def one(cls: type[T], **kwargs: Any) -> SingleQuery[T]:
        ...

    @classmethod
    def id(cls: type[T], id: Union[str, ObjectId]) -> IDQuery[T]:
        ...

    @classmethod
    def linked(cls: type[T]) -> BaseQuery[T]:
        ...

    @classmethod
    def exist(cls: type[T], **kwargs: Any) -> ExistQuery[T]:
        ...

    @classmethod
    def iterate(cls: type[T], **kwargs: Any) -> IterateQuery[T]:
        ...

    def _orm_delete(self: T, no_raise: bool = False) -> None:
        ...

    def _orm_restore(self: T) -> None:
        ...



from __future__ import annotations
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import pymongo, BaseObject


@pymongo
@jsonclass(class_graph='linked')
class LinkedProfile(BaseObject):
    name: str
    user: LinkedUser = types.instanceof('LinkedUser').linkto


@pymongo
@jsonclass(class_graph='linked')
class LinkedUser(BaseObject):
    name: str
    profile: LinkedProfile = types.instanceof('LinkedProfile').linkedby('user')


@pymongo
@jsonclass(class_graph='linked')
class LinkedDProfile(BaseObject):
    name: str
    user: LinkedCUser = types.instanceof('LinkedCUser').linkto.deny


@pymongo
@jsonclass(class_graph='linked')
class LinkedCUser(BaseObject):
    name: str
    profile: LinkedDProfile = types.instanceof('LinkedDProfile') \
                                   .linkedby('user').cascade


@pymongo
@jsonclass(class_graph='linked')
class LinkedCProfile(BaseObject):
    name: str
    user: LinkedDUser = types.instanceof('LinkedDUser').linkto.cascade


@pymongo
@jsonclass(class_graph='linked')
class LinkedDUser(BaseObject):
    name: str
    profile: LinkedCProfile = types.instanceof('LinkedCProfile') \
                                   .linkedby('user').deny



from __future__ import annotations
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import pymongo, BaseObject


@pymongo
@jsonclass(class_graph='linked')
class LinkedCompany(BaseObject):
    name: str
    owners: list[LinkedOwner] = types.listof('LinkedOwner') \
                                     .linkedthru('companies').cascade


@pymongo
@jsonclass(class_graph='linked')
class LinkedOwner(BaseObject):
    name: str
    companies: list[LinkedCompany] = types.listof('LinkedCompany') \
                                          .linkedthru('owners').deny



from __future__ import annotations
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import pymongo, BaseObject


@pymongo
@jsonclass(class_graph='linked')
class LinkedBomb(BaseObject):
    name: str
    soldiers: list[LinkedSoldier] = types.listof('LinkedSoldier') \
                                         .linkedthru('bombs').cascade


@pymongo
@jsonclass(class_graph='linked')
class LinkedSoldier(BaseObject):
    name: str
    bombs: list[LinkedBomb] = types.listof('LinkedBomb') \
                                   .linkedthru('soldiers').cascade



from __future__ import annotations
from unittest import TestCase
from jsonclasses_pymongo import Connection
from tests.classes.simple_song import SimpleSong
from tests.classes.simple_artist import SimpleArtist
from tests.classes.linked_author import LinkedAuthor
from tests.classes.linked_post import LinkedPost
from tests.classes.linked_profile_user import LinkedProfile, LinkedUser
from tests.classes.linked_favorite import LinkedCourse, LinkedStudent
from tests.classes.chained_user import (ChainedAddress, ChainedProfile,
                                        ChainedUser)


class TestSave(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        connection = Connection('simple')
        connection.set_url('mongodb://localhost:27017/simple')
        connection.connect()
        connection = Connection('linked')
        connection.set_url('mongodb://localhost:27017/linked')
        connection.connect()
        connection = Connection('chained')
        connection.set_url('mongodb://localhost:27017/chained')
        connection.connect()

    @classmethod
    def tearDownClass(cls) -> None:
        connection = Connection('simple')
        connection.disconnect()
        connection = Connection('linked')
        connection.disconnect()
        connection = Connection('chained')
        connection.disconnect()

    def setUp(self) -> None:
        collection = Connection.get_collection(SimpleSong)
        collection.delete_many({})
        collection = Connection.get_collection(SimpleArtist)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedAuthor)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedPost)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedProfile)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedUser)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedStudent)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedCourse)
        collection.delete_many({})
        collection = Connection('linked').collection('linkedcoursesstudents'
                                                     'linkedstudentscourses')
        collection.delete_many({})
        collection = Connection.get_collection(ChainedUser)
        collection.delete_many({})
        collection = Connection.get_collection(ChainedProfile)
        collection.delete_many({})
        collection = Connection.get_collection(ChainedAddress)
        collection.delete_many({})

    def test_1f_1l_ref_lookup_fetches_linked_object(self):
        user = LinkedUser(name='Teo Yeo')
        profile = LinkedProfile(name='Great City')
        user.profile = profile
        user.save()
        user = LinkedUser.id(user.id).include('profile').exec()
        self.assertEqual(user.profile.id, profile.id)
        self.assertEqual(user.profile.name, profile.name)

    def test_1l_1f_ref_lookup_fetches_linked_object(self):
        user = LinkedUser(name='Teo Yeo')
        profile = LinkedProfile(name='Great City')
        user.profile = profile
        user.save()
        profile = LinkedProfile.id(profile.id).include('user').exec()
        self.assertEqual(profile.user.id, user.id)
        self.assertEqual(profile.user.name, user.name)

    def test_1f_1l_find_lookup_fetches_linked_object(self):
        user1 = LinkedUser(name='Phou Neng')
        profile1 = LinkedProfile(name='Great City, Too')
        user1.profile = profile1
        user1.save()
        user2 = LinkedUser(name='Teo Yeo')
        profile2 = LinkedProfile(name='Great City')
        user2.profile = profile2
        user2.save()
        results = LinkedUser.find().include('profile').exec()
        self.assertEqual(results[0].id, user1.id)
        self.assertEqual(results[0].profile.id, profile1.id)
        self.assertEqual(results[1].id, user2.id)
        self.assertEqual(results[1].profile.id, profile2.id)

    def test_1l_1f_find_lookup_fetches_linked_object(self):
        user1 = LinkedUser(name='Phou Neng')
        profile1 = LinkedProfile(name='Great City, Too')
        user1.profile = profile1
        user1.save()
        user2 = LinkedUser(name='Teo Yeo')
        profile2 = LinkedProfile(name='Great City')
        user2.profile = profile2
        user2.save()
        results = LinkedProfile.find().include('user').exec()
        self.assertEqual(results[0].id, profile1.id)
        self.assertEqual(results[0].user.id, user1.id)
        self.assertEqual(results[1].id, profile2.id)
        self.assertEqual(results[1].user.id, user2.id)

    def test_1_many_ref_lookup_fetches_linked_objects(self):
        author = LinkedAuthor(name='Kai Nang Piang Tshung')
        post1 = LinkedPost(title='P1', content='A')
        post2 = LinkedPost(title='P2', content='A')
        post3 = LinkedPost(title='P3', content='A')
        post4 = LinkedPost(title='B1', content='A')
        post5 = LinkedPost(title='B2', content='A')
        post6 = LinkedPost(title='B3', content='A')
        author.posts = [post1, post2, post3, post4, post5, post6]
        author.save()
        author = LinkedAuthor(name='Khui Tsia Koi Tsuang Kai Low Rider')
        post1 = LinkedPost(title='P1', content='B')
        post2 = LinkedPost(title='P2', content='B')
        post3 = LinkedPost(title='P3', content='B')
        post4 = LinkedPost(title='B1', content='B')
        post5 = LinkedPost(title='B2', content='B')
        post6 = LinkedPost(title='B3', content='B')
        author.posts = [post1, post2, post3, post4, post5, post6]
        author.save()
        results = LinkedAuthor.find() \
                              .include('posts',
                                       LinkedPost.find(title={'$regex': 'B'}))\
                              .exec()
        self.assertEqual(len(results), 2)
        self.assertEqual(len(results[0].posts), 3)
        self.assertEqual(len(results[1].posts), 3)
        self.assertEqual(results[0].posts[0].title, 'B1')
        self.assertEqual(results[0].posts[1].title, 'B2')
        self.assertEqual(results[0].posts[2].title, 'B3')
        self.assertEqual(results[1].posts[0].title, 'B1')
        self.assertEqual(results[1].posts[1].title, 'B2')
        self.assertEqual(results[1].posts[2].title, 'B3')

    def test_many_1_ref_lookup_fetches_linked_object(self):
        author = LinkedAuthor(name='A1')
        post1 = LinkedPost(title='P1', content='A')
        post2 = LinkedPost(title='P2', content='A')
        post3 = LinkedPost(title='P3', content='A')
        post4 = LinkedPost(title='B1', content='A')
        post5 = LinkedPost(title='B2', content='A')
        post6 = LinkedPost(title='B3', content='A')
        author.posts = [post1, post2, post3, post4, post5, post6]
        author.save()
        author = LinkedAuthor(name='A2')
        post1 = LinkedPost(title='P1', content='B')
        post2 = LinkedPost(title='P2', content='B')
        post3 = LinkedPost(title='P3', content='B')
        post4 = LinkedPost(title='B1', content='B')
        post5 = LinkedPost(title='B2', content='B')
        post6 = LinkedPost(title='B3', content='B')
        author.posts = [post1, post2, post3, post4, post5, post6]
        author.save()
        results = LinkedPost.find(title='P1').include('author').exec()
        self.assertEqual(results[0].author.name, 'A1')
        self.assertEqual(results[1].author.name, 'A2')

    def test_many_many_ref_lookup_fetches_linked_objects(self):
        student1 = LinkedStudent(name='S1')
        student2 = LinkedStudent(name='S2')
        course1 = LinkedCourse(name='C1')
        course2 = LinkedCourse(name='C2')
        student1.courses = [course1, course2]
        student2.courses = [course1, course2]
        student1.save()
        students = LinkedStudent.find().include('courses').exec()
        self.assertEqual(len(students), 2)
        self.assertEqual(len(students[0].courses), 2)
        self.assertEqual(len(students[1].courses), 2)
        self.assertEqual(students[0].id, student2.id)
        self.assertEqual(students[1].id, student1.id)
        self.assertEqual(students[0].courses[0].id, course2.id)
        self.assertEqual(students[0].courses[1].id, course1.id)
        self.assertEqual(students[1].courses[0].id, course2.id)
        self.assertEqual(students[1].courses[1].id, course1.id)

    def test_many_many_ref_lookup_can_be_filtered(self):
        student1 = LinkedStudent(name='S1')
        student2 = LinkedStudent(name='S2')
        course1 = LinkedCourse(name='C1')
        course2 = LinkedCourse(name='C2')
        student1.courses = [course1, course2]
        student2.courses = [course1, course2]
        student1.save()
        students = LinkedStudent.find().include('courses', LinkedCourse.find(name='C1')).exec()
        self.assertEqual(len(students), 2)
        self.assertEqual(len(students[0].courses), 1)
        self.assertEqual(len(students[1].courses), 1)
        self.assertEqual(students[0].courses[0].name, course1.name)
        self.assertEqual(students[1].courses[0].name, course1.name)

    def test_1_1_ref_can_be_chained(self):
        u = ChainedUser(name='U', address={'name': 'A'}, profile={'name': 'P'})
        u.save()
        p = ChainedProfile.id(u.profile.id).include(
            'user', ChainedUser.linked().include('address')).exec()
        self.assertEqual(p.id, u.profile.id)
        self.assertEqual(p.user.id, u.id)
        self.assertEqual(p.user.address.id, u.address.id)

    def test_1_many_ref_uses_unique_objects(self):
        post1 = LinkedPost(title='P1', content='A')
        post2 = LinkedPost(title='P2', content='A')
        author = LinkedAuthor(name='A1', posts=[post1, post2])
        author.save()
        posts = LinkedPost.find().include('author').exec()
        p1 = posts[0]
        p2 = posts[1]
        self.assertIs(p1.author, p2.author)
        p1.author.name = 'New Name'
        p1.save()
        self.assertEqual(p2.author.name, 'New Name')

    def test_many_many_ref_uses_unique_objects(self):
        student1 = LinkedStudent(name='S1')
        student2 = LinkedStudent(name='S2')
        course1 = LinkedCourse(name='C1')
        course2 = LinkedCourse(name='C2')
        student1.courses = [course1, course2]
        student2.courses = [course1, course2]
        student1.save()
        students = LinkedStudent.find().include('courses', LinkedCourse.find(name='C1')).exec()
        s1 = students[0]
        s2 = students[1]
        self.assertIs(s1.courses[0], s2.courses[0])

    def test_fetched_objects_are_unmodified_1f_1l(self):
        user = LinkedUser(name='Bo Hang')
        profile = LinkedProfile(name='Ko Leng')
        user.profile = profile
        user.save()
        profile = LinkedProfile.id(profile.id).include('user').exec()
        self.assertEqual(profile.is_new, False)
        self.assertEqual(profile.is_modified, False)
        self.assertEqual(profile.modified_fields, ())
        self.assertEqual(profile.user.is_new, False)
        self.assertEqual(profile.user.is_modified, False)
        self.assertEqual(profile.user.modified_fields, ())

    def test_fetched_objects_are_unmodified_1l_1f(self):
        user = LinkedUser(name='Bo Hang')
        profile = LinkedProfile(name='Ko Leng')
        user.profile = profile
        user.save()
        user = LinkedUser.id(user.id).include('profile').exec()
        self.assertEqual(user.is_new, False)
        self.assertEqual(user.is_modified, False)
        self.assertEqual(user.modified_fields, ())
        self.assertEqual(user.profile.is_new, False)
        self.assertEqual(user.profile.is_modified, False)
        self.assertEqual(user.profile.modified_fields, ())

    def test_fetched_objects_are_unmodified_1_many(self):
        author = LinkedAuthor(name='Me')
        post1 = LinkedPost(title='P1', content='A')
        post2 = LinkedPost(title='P2', content='A')
        author.posts = [post1, post2]
        author.save()
        authors = LinkedAuthor.find().include('posts').exec()
        a = authors[0]
        p1 = a.posts[0]
        p2 = a.posts[1]
        self.assertEqual(a.is_new, False)
        self.assertEqual(a.is_modified, False)
        self.assertEqual(a.modified_fields, ())
        self.assertEqual(p1.is_new, False)
        self.assertEqual(p1.is_modified, False)
        self.assertEqual(p1.modified_fields, ())
        self.assertEqual(p2.is_new, False)
        self.assertEqual(p2.is_modified, False)
        self.assertEqual(p2.modified_fields, ())

    def test_fetched_objects_are_unmodified_many_1(self):
        author = LinkedAuthor(name='Me')
        post1 = LinkedPost(title='P1', content='A')
        post2 = LinkedPost(title='P2', content='A')
        author.posts = [post1, post2]
        author.save()
        posts = LinkedPost.find().include('author').exec()
        self.assertEqual(posts[0].is_new, False)
        self.assertEqual(posts[0].is_modified, False)
        self.assertEqual(posts[0].modified_fields, ())
        self.assertEqual(posts[1].is_new, False)
        self.assertEqual(posts[1].is_modified, False)
        self.assertEqual(posts[1].modified_fields, ())
        self.assertEqual(posts[0].author.is_new, False)
        self.assertEqual(posts[0].author.is_modified, False)
        self.assertEqual(posts[0].author.modified_fields, ())
        self.assertEqual(posts[0].author, posts[1].author)

    def test_fetched_objects_are_unmodified_many_many(self):
        student1 = LinkedStudent(name='S1')
        student2 = LinkedStudent(name='S2')
        course1 = LinkedCourse(name='C1')
        course2 = LinkedCourse(name='C2')
        student1.courses = [course1, course2]
        student2.courses = [course1, course2]
        student1.save()
        students = LinkedStudent.find().include('courses').exec()
        self.assertEqual(students[0].is_new, False)
        self.assertEqual(students[0].is_modified, False)
        self.assertEqual(students[0].modified_fields, ())
        self.assertEqual(students[1].is_new, False)
        self.assertEqual(students[1].is_modified, False)
        self.assertEqual(students[1].modified_fields, ())
        self.assertEqual(students[0].courses[0].is_new, False)
        self.assertEqual(students[0].courses[0].is_modified, False)
        self.assertEqual(students[0].courses[0].modified_fields, ())
        self.assertEqual(students[0].courses[1].is_new, False)
        self.assertEqual(students[0].courses[1].is_modified, False)
        self.assertEqual(students[0].courses[1].modified_fields, ())
        self.assertEqual(students[1].courses[0].is_new, False)
        self.assertEqual(students[1].courses[0].is_modified, False)
        self.assertEqual(students[1].courses[0].modified_fields, ())
        self.assertEqual(students[1].courses[1].is_new, False)
        self.assertEqual(students[1].courses[1].is_modified, False)
        self.assertEqual(students[1].courses[1].modified_fields, ())



from __future__ import annotations
from typing import Any
from pymongo.collection import Collection


class Command:

    def execute(self) -> None:
        raise NotImplementedError(
            'Please use concrete subclasses of Command.')

    def __repr__(self) -> str:
        return '<Command()>'


class InsertOneCommand(Command):

    def __init__(self, collection: Collection, object: dict[str, Any]) -> None:
        self.collection = collection
        self.object = object

    def execute(self) -> None:
        self.collection.insert_one(self.object)

    def __repr__(self) -> str:
        return (f'<InsertOneCommand(collection={self.collection.name}, '
                f'object={self.object})>')


class UpdateOneCommand(Command):

    def __init__(self,
                 collection: Collection,
                 object: dict[str, Any],
                 matcher: dict[str, Any]) -> None:
        self.collection = collection
        self.object = object
        self.matcher = matcher
        self.upsert = False

    def execute(self) -> None:
        self.collection.update_one(
            filter=self.matcher,
            update=self.object,
            upsert=self.upsert)

    def __repr__(self) -> str:
        return (f'<UpdateOneCommand(collection={self.collection.name}, '
                f'object={self.object}), matcher={self.matcher}>')


class UpsertOneCommand(UpdateOneCommand):

    def __init__(self,
                 collection: Collection,
                 object: dict[str, Any],
                 matcher: dict[str, Any]) -> None:
        super().__init__(collection=collection, object=object, matcher=matcher)
        self.upsert = True

    def __repr__(self) -> str:
        return (f'<UpsertOneCommand(collection={self.collection.name}, '
                f'object={self.object}), matcher={self.matcher}>')


class DeleteOneCommand(Command):

    def __init__(self,
                 collection: Collection,
                 matcher: dict[str, Any]) -> None:
        self.collection = collection
        self.matcher = matcher

    def execute(self) -> None:
        self.collection.delete_one(filter=self.matcher)

    def __repr__(self) -> str:
        return (f'<UpsertOneCommand(collection={self.collection.name}, '
                f'matcher={self.matcher}>')


class BatchCommand(Command):

    def __init__(self, commands: list[Command]) -> None:
        self.commands = commands

    def execute(self) -> None:
        for command in self.commands:
            command.execute()



from __future__ import annotations
from typing import TYPE_CHECKING
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import pymongo, BaseObject
if TYPE_CHECKING:
    from .linked_author import LinkedAuthor


@pymongo
@jsonclass(class_graph='linked')
class LinkedPost(BaseObject):
    title: str
    content: str
    author: LinkedAuthor = types.linkto.instanceof('LinkedAuthor')



from __future__ import annotations
from typing import Any, NamedTuple, TypeVar, Union, cast, TYPE_CHECKING
from datetime import datetime
from inflection import camelize
from bson.objectid import ObjectId
from jsonclasses.jsonclass_field import JSONClassField
from jsonclasses.field_definition import FieldStorage, FieldType
from jsonclasses.keypath_utils import concat_keypath
from jsonclasses.types_resolver import TypesResolver
from jsonclasses.mark_graph import MarkGraph
from jsonclasses.types import types
from .coder import Coder
from .utils import ref_db_field_key, ref_db_field_keys, ref_field_key
from .context import EncodingContext
from .command import (Command, InsertOneCommand, UpdateOneCommand,
                      UpsertOneCommand, DeleteOneCommand, BatchCommand)
from .connection import Connection

if TYPE_CHECKING:
    from .pymongo_object import PymongoObject
    T = TypeVar('T', bound=PymongoObject)


class EncodingResult(NamedTuple):
    """The result from encoding an item."""
    result: Any
    commands: list[Command]


class Encoder(Coder):
    """Write commands encoder."""

    def encode_list(self, context: EncodingContext) -> EncodingResult:
        if context.value is None:
            return EncodingResult(result=None, commands=[])
        value = cast(list[Any], context.value)
        fd = context.types.definition
        item_types = TypesResolver().resolve_types(
            fd.raw_item_types, context.root.__class__.definition.config)
        if fd.field_storage == FieldStorage.FOREIGN_KEY:
            item_types = item_types.linkedby(cast(str, fd.foreign_key))
        if fd.field_storage == FieldStorage.LOCAL_KEY:
            item_types = item_types.linkto
        result = []
        commands = []
        for index, item in enumerate(value):
            item_result, item_commands = self.encode_item(context.new(
                value=item,
                types=item_types,
                keypath_root=concat_keypath(context.keypath_root, index),
                keypath_owner=concat_keypath(context.keypath_owner, index),
                keypath_parent=str(index),
                parent=value))
            result.append(item_result)
            commands.extend(item_commands)
        return EncodingResult(result, commands)

    def encode_dict(self, context: EncodingContext) -> EncodingResult:
        if context.value is None:
            return EncodingResult(result=None, commands=[])
        value = cast(dict[str, Any], context.value)
        fd = context.types.definition
        item_types = TypesResolver().resolve_types(fd.raw_item_types)
        camelized = context.owner.__class__.dbconf.camelize_db_keys
        result = {}
        commands = []
        for key, item in value.items():
            item_result, item_commands = self.encode_item(context.new(
                value=item,
                types=item_types,
                keypath_root=concat_keypath(context.keypath_root, key),
                keypath_owner=concat_keypath(context.keypath_owner, key),
                keypath_parent=str(key),
                parent=value))
            result[camelize(key, False) if camelized else key] = item_result
            commands.extend(item_commands)
        return EncodingResult(result, commands)

    def encode_shape(self, context: EncodingContext) -> EncodingResult:
        if context.value is None:
            return EncodingResult(result=None, commands=[])
        value = cast(dict[str, Any], context.value)
        fd = context.types.definition
        shape_types = cast(dict[str, Any], fd.shape_types)
        camelized = context.owner.__class__.dbconf.camelize_db_keys
        result = {}
        commands = []
        for key, item in value.items():
            item_types = TypesResolver().resolve_types(shape_types[key])
            item_result, item_commands = self.encode_item(context.new(
                value=item,
                types=item_types,
                keypath_root=concat_keypath(context.keypath_root, key),
                keypath_owner=concat_keypath(context.keypath_owner, key),
                keypath_parent=str(key),
                parent=value))
            result[camelize(key, False) if camelized else key] = item_result
            commands.extend(item_commands)
        return EncodingResult(result, commands)

    def _join_command(self,
                      this_instance: PymongoObject,
                      this_field: JSONClassField,
                      that_cls: type[PymongoObject],
                      that_id: ObjectId) -> UpsertOneCommand:
        this_cls = this_instance.__class__
        connection = Connection.from_class(this_instance.__class__)
        this_cls_name = this_cls.__name__
        that_cls_name = that_cls.__name__
        this_field_name = ref_db_field_key(this_cls_name, this_cls)
        that_field_name = ref_db_field_key(that_cls_name, that_cls)
        join_table_name = self.join_table_name(
            this_cls,
            this_field.name,
            that_cls,
            cast(str, this_field.definition.foreign_key))
        collection = connection.collection(join_table_name)
        this_id = ObjectId(this_instance._id)
        matcher = {
            this_field_name: this_id,
            that_field_name: that_id
        }
        return UpsertOneCommand(collection=collection,
                                object={'$set': matcher},
                                matcher=matcher)

    def _unlink_command(self,
                        this_instance: PymongoObject,
                        this_field: JSONClassField,
                        that_cls: type[PymongoObject],
                        that_id: ObjectId) -> DeleteOneCommand:
        this_cls = this_instance.__class__
        connection = Connection.from_class(this_instance.__class__)
        this_cls_name = this_cls.__name__
        that_cls_name = that_cls.__name__
        this_field_name = ref_db_field_key(this_cls_name, this_cls)
        that_field_name = ref_db_field_key(that_cls_name, that_cls)
        join_table_name = self.join_table_name(
            this_cls,
            this_field.name,
            that_cls,
            cast(str, this_field.definition.foreign_key))
        collection = connection.collection(join_table_name)
        this_id = ObjectId(this_instance._id)
        matcher = {
            this_field_name: this_id,
            that_field_name: that_id
        }
        return DeleteOneCommand(collection=collection, matcher=matcher)

    def encode_instance(self,
                        context: EncodingContext,
                        root: bool = False) -> EncodingResult:
        from .pymongo_object import PymongoObject
        if context.value is None:
            return EncodingResult(result=None, commands=[])
        value = cast(PymongoObject, context.value)
        cls = value.__class__
        id = cast(Union[str, int], value._id)
        if context.mark_graph.getp(cls, id) is not None:
            return EncodingResult({'_id': ObjectId(id)}, commands=[])
        context.mark_graph.put(value)
        instance_fd = context.types.definition
        write_instance = instance_fd.field_storage != FieldStorage.EMBEDDED
        if root:
            write_instance = True
        use_insert_command = False
        fields_need_update: set[str] = value.persisted_modified_fields
        if value.is_new:
            use_insert_command = True
        result_set = {}
        matcher = {}
        commands = []
        result_addtoset = {}
        for field in value.__class__.definition.fields:
            if field.definition.is_temp_field:
                continue
            fname = field.name
            fvalue = getattr(value, fname)
            ftypes = field.types
            if self.is_id_field(field):
                if use_insert_command:
                    result_set['_id'] = ObjectId(fvalue)
                else:
                    matcher['_id'] = ObjectId(fvalue)
            elif self.is_foreign_key_reference_field(field):
                if fvalue is None:
                    continue
                _, item_commands = self.encode_instance(context.new(
                    value=fvalue,
                    types=ftypes,
                    keypath_root=concat_keypath(context.keypath_root, fname),
                    keypath_owner=fname,
                    owner=value,
                    keypath_parent=fname,
                    parent=value))
                commands.extend(item_commands)
            elif self.is_foreign_keys_reference_field(field):
                if fvalue is None:
                    continue
                item_result, item_commands = self.encode_list(context.new(
                    value=fvalue,
                    types=ftypes,
                    keypath_root=concat_keypath(context.keypath_root, fname),
                    keypath_owner=fname,
                    owner=value,
                    keypath_parent=fname,
                    parent=value))
                commands.extend(item_commands)
                if not self.is_join_table_field(field):
                    continue
                for list_item in item_result:
                    if list_item.get('_id'):
                        join_command = self._join_command(
                            value,
                            field,
                            self.list_instance_type(field, cls),
                            list_item['_id'])
                        commands.append(join_command)
                for item in value.unlinked_objects[field.name]:
                    if item._id is not None:
                        unlink_command = self._unlink_command(
                            value,
                            field,
                            self.list_instance_type(field, cls),
                            ObjectId(item._id))
                        commands.append(unlink_command)
            elif self.is_local_key_reference_field(field):
                if fvalue is None:
                    tsfm = value.__class__.definition.config.key_transformer
                    if getattr(value, tsfm(field)) is not None:
                        if use_insert_command or fname in fields_need_update:
                            result_set[ref_db_field_key(fname, cls)] = \
                                ObjectId(getattr(value, tsfm(field)))
                    else:
                        if use_insert_command or fname in fields_need_update:
                            result_set[ref_db_field_key(fname, cls)] = None
                    continue
                item_result, item_commands = self.encode_instance(context.new(
                    value=fvalue,
                    types=ftypes,
                    keypath_root=concat_keypath(context.keypath_root, fname),
                    keypath_owner=fname,
                    owner=value,
                    keypath_parent=fname,
                    parent=value))
                if use_insert_command or fname in fields_need_update:
                    fname_ref = ref_db_field_key(fname, cls)
                    result_set[fname_ref] = item_result['_id']
                    setattr(value, ref_field_key(fname),
                                   str(item_result['_id']))
                commands.extend(item_commands)
            elif self.is_local_keys_reference_field(field):
                if fvalue is None:
                    if use_insert_command or fname in fields_need_update:
                        result_set[ref_db_field_keys(fname, cls)] = None
                    continue
                item_result, item_commands = self.encode_list(context.new(
                    value=fvalue,
                    types=ftypes,
                    keypath_root=concat_keypath(context.keypath_root, fname),
                    keypath_owner=fname,
                    owner=value,
                    keypath_parent=fname,
                    parent=value))
                if use_insert_command or fname in fields_need_update:
                    id_list = [result['_id'] for result in item_result]
                    fname_ref = ref_db_field_keys(fname, cls)
                    if use_insert_command:
                        result_set[fname_ref] = id_list
                    else:
                        result_addtoset[fname_ref] = {'$each': id_list}
                commands.extend(item_commands)
            else:
                item_result, item_commands = self.encode_item(context.new(
                    value=fvalue,
                    types=ftypes,
                    keypath_root=concat_keypath(context.keypath_root, fname),
                    keypath_owner=concat_keypath(context.keypath_owner, fname),
                    owner=value,
                    keypath_parent=fname,
                    parent=value))
                if use_insert_command or fname in fields_need_update:
                    if context.owner.__class__.dbconf.camelize_db_keys:
                        result_set[camelize(field.name, False)] = item_result
                    else:
                        result_set[field.name] = item_result
                commands.extend(item_commands)
        if write_instance:
            collection = Connection.get_collection(value.__class__)
            if use_insert_command:
                insert_command = InsertOneCommand(collection, result_set)
                commands.append(insert_command)
            else:
                updator = {}
                if len(result_set) > 0:
                    updator['$set'] = result_set
                if len(result_addtoset) > 0:
                    updator['$addToSet'] = result_addtoset
                if len(updator) > 0:
                    update_c = UpdateOneCommand(collection, updator, matcher)
                    commands.append(update_c)
        value._clear_temp_fields()
        setattr(value, '_is_new', False)
        setattr(value, '_is_modified', False)
        setattr(value, '_modified_fields', set())
        setattr(value, '_previous_values', {})
        if result_set == {}:
            result_set = {'_id': ObjectId(value._id)}
        return EncodingResult(result_set, commands)

    def encode_item(self, context: EncodingContext) -> EncodingResult:
        if context.value is None:
            return EncodingResult(result=None, commands=[])
        field_type = context.types.definition.field_type
        if field_type == FieldType.LIST:
            return self.encode_list(context)
        elif field_type == FieldType.DICT:
            return self.encode_dict(context)
        elif field_type == FieldType.SHAPE:
            return self.encode_shape(context)
        elif field_type == FieldType.INSTANCE:
            return self.encode_instance(context)
        elif field_type == FieldType.DATE:
            return EncodingResult(
                result=datetime.fromisoformat(context.value.isoformat()),
                commands=[])
        elif field_type == FieldType.ENUM:
            return EncodingResult(result=context.value.value, commands=[])
        else:
            return EncodingResult(context.value, [])

    def encode_root(self, root: T) -> BatchCommand:
        commands = self.encode_instance(EncodingContext(
            value=root,
            types=types.instanceof(root.__class__),
            keypath_root='',
            root=root,
            keypath_owner='',
            owner=root,
            keypath_parent='',
            parent=root,
            mark_graph=MarkGraph()), root=True)[1]
        return BatchCommand(commands=commands)



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import pymongo, BaseObject


@pymongo
@jsonclass(class_graph='simple')
class SimpleSinger(BaseObject):
    name: Optional[str] = types.str.unique



from __future__ import annotations
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import pymongo, BaseObject


@pymongo
@jsonclass(class_graph='chained')
class ChainedProfile(BaseObject):
    name: str
    user: ChainedUser = types.instanceof('ChainedUser').linkto


@pymongo
@jsonclass(class_graph='chained')
class ChainedUser(BaseObject):
    name: str
    profile: ChainedProfile = types.instanceof('ChainedProfile').linkedby('user')
    address: ChainedAddress = types.instanceof('ChainedAddress').linkedby('user')


@pymongo
@jsonclass(class_graph='chained')
class ChainedAddress(BaseObject):
    name: str
    user: ChainedUser = types.instanceof('ChainedUser').linkto



from __future__ import annotations
from jsonclasses_pymongo.connection import Connection
from unittest import TestCase
from typing import List, Dict, cast
from datetime import date, datetime
from bson import ObjectId
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import BaseObject, pymongo
from jsonclasses_pymongo.encoder import Encoder
from jsonclasses_pymongo.command import InsertOneCommand


class TestEncoder(TestCase):

    def test_encode_str_into_str(self):
        @pymongo
        @jsonclass
        class SimpleEncodeStr(BaseObject):
            val1: str
            val2: str
        simple_object = SimpleEncodeStr(val1='q', val2='e')
        batch_command = Encoder().encode_root(simple_object)
        insert_command = cast(InsertOneCommand, batch_command.commands[0])
        serialized = insert_command.object
        self.assertEqual(set(serialized.keys()), set(
            ['_id', 'createdAt', 'updatedAt', 'val1', 'val2', 'deletedAt']))
        self.assertIsInstance(serialized['_id'], ObjectId)
        self.assertEqual(serialized['val1'], 'q')
        self.assertEqual(serialized['val2'], 'e')

    def test_encode_int_into_int(self):
        @pymongo
        @jsonclass
        class SimpleEncodeInt(BaseObject):
            age: int
            length: int
        simple_object = SimpleEncodeInt(age=4, length=8)
        batch_command = Encoder().encode_root(simple_object)
        insert_command = cast(InsertOneCommand, batch_command.commands[0])
        serialized = insert_command.object
        self.assertEqual(set(serialized.keys()), set(
            ['_id', 'createdAt', 'updatedAt', 'age', 'length', 'deletedAt']))
        self.assertIsInstance(serialized['_id'], ObjectId)
        self.assertEqual(serialized['age'], 4)
        self.assertEqual(serialized['length'], 8)

    def test_encode_float_into_float(self):
        @pymongo
        @jsonclass
        class SimpleEncodeFloat(BaseObject):
            width: float
            length: float
        simple_object = SimpleEncodeFloat(width=4.5, length=8.5)
        batch_command = Encoder().encode_root(simple_object)
        insert_command = cast(InsertOneCommand, batch_command.commands[0])
        serialized = insert_command.object
        self.assertEqual(set(serialized.keys()), set(
            ['_id', 'createdAt', 'updatedAt', 'width', 'length', 'deletedAt']))
        self.assertIsInstance(serialized['_id'], ObjectId)
        self.assertEqual(serialized['width'], 4.5)
        self.assertEqual(serialized['length'], 8.5)

    def test_encode_bool_into_bool(self):
        @pymongo
        @jsonclass
        class SimpleEncodeBool(BaseObject):
            b1: bool
            b2: bool
        simple_object = SimpleEncodeBool(b1=True, b2=False)
        batch_command = Encoder().encode_root(simple_object)
        insert_command = cast(InsertOneCommand, batch_command.commands[0])
        serialized = insert_command.object
        self.assertEqual(set(serialized.keys()), set(
            ['_id', 'createdAt', 'updatedAt', 'b1', 'b2', 'deletedAt']))
        self.assertIsInstance(serialized['_id'], ObjectId)
        self.assertEqual(serialized['b1'], True)
        self.assertEqual(serialized['b2'], False)

    def test_encode_date_into_datetime(self):
        @pymongo
        @jsonclass
        class SimpleEncodeDate(BaseObject):
            d1: date = date(2012, 9, 15)
            d2: date = date(2020, 9, 14)
        simple_object = SimpleEncodeDate()
        batch_command = Encoder().encode_root(simple_object)
        insert_command = cast(InsertOneCommand, batch_command.commands[0])
        serialized = insert_command.object
        self.assertEqual(set(serialized.keys()), set(
            ['_id', 'createdAt', 'updatedAt', 'd1', 'd2', 'deletedAt']))
        self.assertIsInstance(serialized['_id'], ObjectId)
        self.assertEqual(serialized['d1'], datetime(2012, 9, 15, 0, 0, 0))
        self.assertEqual(serialized['d2'], datetime(2020, 9, 14, 0, 0, 0))

    def test_encode_datetime_into_datetime(self):
        @pymongo
        @jsonclass
        class SimpleEncodeDatetime(BaseObject):
            d1: date = datetime(2012, 9, 15, 0, 0, 0)
            d2: date = datetime(2020, 9, 14, 0, 0, 0)
        simple_object = SimpleEncodeDatetime()
        batch_command = Encoder().encode_root(simple_object)
        insert_command = cast(InsertOneCommand, batch_command.commands[0])
        serialized = insert_command.object
        self.assertEqual(set(serialized.keys()), set(
            ['_id', 'createdAt', 'updatedAt', 'd1', 'd2', 'deletedAt']))
        self.assertIsInstance(serialized['_id'], ObjectId)
        self.assertEqual(serialized['d1'], datetime(2012, 9, 15, 0, 0, 0))
        self.assertEqual(serialized['d2'], datetime(2020, 9, 14, 0, 0, 0))

    def test_encode_embedded_list(self):
        @pymongo
        @jsonclass
        class SimpleEncodeScalarTypesList(BaseObject):
            str_values: List[str]
            int_values: List[int]
            bool_values: List[bool]
        simple_object = SimpleEncodeScalarTypesList(str_values=['0', '1'],
                                                    int_values=[0, 1],
                                                    bool_values=[False, True])
        batch_command = Encoder().encode_root(simple_object)
        insert_command = cast(InsertOneCommand, batch_command.commands[0])
        serialized = insert_command.object
        self.assertEqual(set(serialized.keys()), set(
            ['_id', 'createdAt', 'updatedAt', 'strValues', 'intValues',
             'boolValues', 'deletedAt']))
        self.assertIsInstance(serialized['_id'], ObjectId)
        self.assertEqual(serialized['strValues'], ['0', '1'])
        self.assertEqual(serialized['intValues'], [0, 1])
        self.assertEqual(serialized['boolValues'], [False, True])

    def test_encode_embedded_dict(self):
        @pymongo
        @jsonclass
        class SimpleEncodeScalarTypesDict(BaseObject):
            str_values: Dict[str, str]
        simple_object = SimpleEncodeScalarTypesDict(
            str_values={'0': 'zero', '1': 'one'})
        batch_command = Encoder().encode_root(simple_object)
        insert_command = cast(InsertOneCommand, batch_command.commands[0])
        serialized = insert_command.object
        self.assertEqual(set(serialized.keys()), set(
            ['_id', 'createdAt', 'updatedAt', 'strValues', 'deletedAt']))
        self.assertIsInstance(serialized['_id'], ObjectId)
        self.assertEqual(serialized['strValues'], {'0': 'zero', '1': 'one'})

    def test_encode_embedded_shape(self):
        @pymongo
        @jsonclass
        class SimpleEncodeScalarTypesShape(BaseObject):
            vals: Dict[str, str] = types.shape({
                '0': str,
                '1': int
            })
        simple_object = SimpleEncodeScalarTypesShape(
            vals={'0': 'zero', '1': 1})
        batch_command = Encoder().encode_root(simple_object)
        insert_command = cast(InsertOneCommand, batch_command.commands[0])
        serialized = insert_command.object
        self.assertEqual(set(serialized.keys()), set(
            ['_id', 'createdAt', 'updatedAt', 'vals', 'deletedAt']))
        self.assertIsInstance(serialized['_id'], ObjectId)
        self.assertEqual(serialized['vals'], {'0': 'zero', '1': 1})

    def test_encode_embedded_instance(self):
        @pymongo
        @jsonclass
        class SimpleEncodeEmbeddedInstanceAddress(BaseObject):
            line1: str

        @pymongo
        @jsonclass
        class SimpleEncodeEmbeddedInstance(BaseObject):
            address: SimpleEncodeEmbeddedInstanceAddress
        simple_object = SimpleEncodeEmbeddedInstance(
            address={'line1': 'Flam Road'})
        batch_command = Encoder().encode_root(simple_object)
        commands = batch_command.commands
        self.assertEqual(len(commands), 1)
        command = commands[0]
        self.assertIs(command.collection,
                      Connection.get_collection(SimpleEncodeEmbeddedInstance))
        data = command.object
        address = data['address']
        self.assertIsInstance(address['_id'], ObjectId)
        self.assertEqual(address['line1'], 'Flam Road')

    def test_encode_foreign_key_instance(self):
        @pymongo
        @jsonclass
        class SimpleEncodeForeignKeyInstanceAddress(BaseObject):
            line1: str
            owner: SimpleEncodeForeignKeyInstance = types.linkto.instanceof(
                'SimpleEncodeForeignKeyInstance')

        @pymongo
        @jsonclass
        class SimpleEncodeForeignKeyInstance(BaseObject):
            address: SimpleEncodeForeignKeyInstanceAddress = types.instanceof(
                SimpleEncodeForeignKeyInstanceAddress).linkedby('owner')
        simple_object = SimpleEncodeForeignKeyInstance(
            address={'line1': 'Flam Road'})
        batch_command = Encoder().encode_root(simple_object)
        commands = batch_command.commands
        self.assertEqual(len(commands), 2)
        address_data = commands[0].object
        owner_data = commands[1].object
        self.assertEqual(address_data['ownerId'], owner_data['_id'])

    def test_encode_local_key_instance(self):
        @pymongo
        @jsonclass
        class SimpleEncodeLocalKeyInstanceAddress(BaseObject):
            line1: str
            owner: SimpleEncodeLocalKeyInstance = types.instanceof(
                'SimpleEncodeLocalKeyInstance').linkedby('address')

        @pymongo
        @jsonclass
        class SimpleEncodeLocalKeyInstance(BaseObject):
            address: SimpleEncodeLocalKeyInstanceAddress = types.linkto.instanceof(SimpleEncodeLocalKeyInstanceAddress)
        simple_object = SimpleEncodeLocalKeyInstance(
            address={'line1': 'Flam Road'})
        batch_command = Encoder().encode_root(simple_object)
        commands = batch_command.commands
        self.assertEqual(len(commands), 2)
        address_data = commands[0].object
        owner_data = commands[1].object
        self.assertEqual(owner_data['addressId'], address_data['_id'])

    def test_encode_embedded_instance_list(self):
        @pymongo
        @jsonclass
        class MediumEncodeEmbeddedInstanceAddress(BaseObject):
            line1: str

        @pymongo
        @jsonclass
        class MediumEncodeEmbeddedInstance(BaseObject):
            addresses: List[MediumEncodeEmbeddedInstanceAddress]
        medium_object = MediumEncodeEmbeddedInstance(
            addresses=[{'line1': 'Flam Road'}, {'line1': 'Plam Road'}])
        batch_command = Encoder().encode_root(medium_object)
        commands = batch_command.commands
        self.assertEqual(len(commands), 1)
        command = commands[0]
        self.assertIs(command.collection,
                      Connection.get_collection(MediumEncodeEmbeddedInstance))
        data = command.object
        addresses = data['addresses']
        self.assertEqual(len(addresses), 2)
        self.assertEqual(addresses[0]['line1'], 'Flam Road')
        self.assertEqual(addresses[1]['line1'], 'Plam Road')

    def test_encode_foreign_keys_instance_list(self):
        @pymongo
        @jsonclass
        class MediumEncodeForeignKeyInstanceAddress(BaseObject):
            line1: str
            owner: MediumEncodeForeignKeyInstance = types.linkto.instanceof(
                'MediumEncodeForeignKeyInstance')

        @pymongo
        @jsonclass
        class MediumEncodeForeignKeyInstance(BaseObject):
            addresses: List[MediumEncodeForeignKeyInstanceAddress] = types.listof(
                MediumEncodeForeignKeyInstanceAddress).linkedby('owner')
        simple_object = MediumEncodeForeignKeyInstance(
            addresses=[{'line1': 'Flam Road'}, {'line1': 'Klam Road'}])
        batch_command = Encoder().encode_root(simple_object)
        commands = batch_command.commands
        self.assertEqual(len(commands), 3)
        address_0_data = commands[0].object
        address_1_data = commands[1].object
        owner_data = commands[2].object
        self.assertEqual(address_0_data['ownerId'], owner_data['_id'])
        self.assertEqual(address_1_data['ownerId'], owner_data['_id'])

    def test_encode_local_keys_instance_list(self):
        @pymongo
        @jsonclass
        class MediumEncodeLocalKeyInstanceAddress(BaseObject):
            line1: str
            owner: MediumEncodeLocalKeyInstance = types.linkedby(
                'address').instanceof('SimpleEncodeLocalKeyInstance')

        @pymongo
        @jsonclass
        class MediumEncodeLocalKeyInstance(BaseObject):
            addresses: List[MediumEncodeLocalKeyInstanceAddress] = types.listof(
                MediumEncodeLocalKeyInstanceAddress).linkto
        simple_object = MediumEncodeLocalKeyInstance(
            addresses=[{'line1': 'Flam Road'}, {'line1': 'Klam Road'}])
        batch_command = Encoder().encode_root(simple_object)
        commands = batch_command.commands
        self.assertEqual(len(commands), 3)
        address_0_data = commands[0].object
        address_1_data = commands[1].object
        owner_data = commands[2].object
        self.assertEqual(owner_data['addressesIds'], [
                         address_0_data['_id'], address_1_data['_id']])

    def test_encode_dict_camelize_keys(self):
        @pymongo
        @jsonclass
        class MediumEncodeCamelizeDictKeys(BaseObject):
            val: Dict[str, str] = types.dictof(types.str)
        simple_object = MediumEncodeCamelizeDictKeys(
            val={'key_one': 'val_one', 'key_two': 'val_two'})
        batch_command = Encoder().encode_root(simple_object)
        commands = batch_command.commands
        encoded_val = commands[0].object['val']
        self.assertEqual(
            encoded_val, {'keyOne': 'val_one', 'keyTwo': 'val_two'})

    def test_encode_dict_do_not_camelize_keys(self):
        @pymongo(camelize_db_keys=False)
        @jsonclass
        class MediumEncodeDoNotCamelizeDictKeys(BaseObject):
            val: Dict[str, str] = types.dictof(types.str)
        simple_object = MediumEncodeDoNotCamelizeDictKeys(
            val={'key_one': 'val_one', 'key_two': 'val_two'})
        batch_command = Encoder().encode_root(simple_object)
        commands = batch_command.commands
        encoded_val = commands[0].object['val']
        self.assertEqual(
            encoded_val, {'key_one': 'val_one', 'key_two': 'val_two'})

    def test_encode_shape_camelize_keys(self):
        @pymongo
        @jsonclass
        class MediumEncodeCamelizeShapeKeys(BaseObject):
            val: Dict[str, str] = types.shape({
                'key_one': types.str,
                'key_two': types.str
            })
        simple_object = MediumEncodeCamelizeShapeKeys(
            val={'key_one': 'val_one', 'key_two': 'val_two'})
        batch_command = Encoder().encode_root(simple_object)
        commands = batch_command.commands
        encoded_val = commands[0].object['val']
        self.assertEqual(
            encoded_val, {'keyOne': 'val_one', 'keyTwo': 'val_two'})

    def test_encode_shape_do_not_camelize_keys(self):
        @pymongo(camelize_db_keys=False)
        @jsonclass
        class MediumEncodeDoNotCamelizeShapeKeys(BaseObject):
            val: Dict[str, str] = types.shape({
                'key_one': types.str,
                'key_two': types.str
            })
        simple_object = MediumEncodeDoNotCamelizeShapeKeys(
            val={'key_one': 'val_one', 'key_two': 'val_two'})
        batch_command = Encoder().encode_root(simple_object)
        commands = batch_command.commands
        encoded_val = commands[0].object['val']
        self.assertEqual(
            encoded_val, {'key_one': 'val_one', 'key_two': 'val_two'})

    def test_encoder_handle_many_to_many_with_linkedthru(self):
        @pymongo
        @jsonclass
        class DifficultEncodeLinkedThruA(BaseObject):
            aval: str
            blinks: List[DifficultEncodeLinkedThruB] = types.listof(
                'DifficultEncodeLinkedThruB').linkedthru('alinks')

        @pymongo
        @jsonclass
        class DifficultEncodeLinkedThruB(BaseObject):
            bval: str
            alinks: List[DifficultEncodeLinkedThruA] = types.listof(
                'DifficultEncodeLinkedThruA').linkedthru('blinks')
        instance_a = DifficultEncodeLinkedThruA(
            aval='A', blinks=[{'bval': 'B1'}, {'bval': 'B2'}])
        batch_command = Encoder().encode_root(instance_a)
        commands = batch_command.commands
        self.assertEqual(len(commands), 7)



from __future__ import annotations
from unittest import TestCase
from jsonclasses.exceptions import DeletionDeniedException
from jsonclasses_pymongo import Connection
from tests.classes.simple_song import SimpleSong
from tests.classes.simple_artist import SimpleArtist
from tests.classes.linked_author import LinkedAuthor
from tests.classes.linked_post import LinkedPost
from tests.classes.linked_profile_user import (LinkedProfile, LinkedUser,
                                               LinkedCProfile, LinkedDUser,
                                               LinkedDProfile, LinkedCUser)
from tests.classes.linked_favorite import LinkedCourse, LinkedStudent
from tests.classes.linked_order import (LinkedOrder, LinkedBuyer, LinkedCOrder,
                                        LinkedCBuyer)
from tests.classes.linked_account import LinkedAccount, LinkedBalance
from tests.classes.linked_bomb import LinkedSoldier, LinkedBomb
from tests.classes.linked_notebook import (LinkedNote, LinkedNotebook,
                                           LinkedRNote, LinkedRNotebook)
from tests.classes.linked_share import LinkedCompany, LinkedOwner


class TestDelete(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        connection = Connection('simple')
        connection.set_url('mongodb://localhost:27017/simple')
        connection.connect()
        connection = Connection('linked')
        connection.set_url('mongodb://localhost:27017/linked')
        connection.connect()

    @classmethod
    def tearDownClass(cls) -> None:
        connection = Connection('simple')
        connection.disconnect()
        connection = Connection('linked')
        connection.disconnect()

    def setUp(self) -> None:
        collection = Connection.get_collection(SimpleSong)
        collection.delete_many({})
        collection = Connection.get_collection(SimpleArtist)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedAuthor)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedPost)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedUser)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedProfile)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedStudent)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedCourse)
        collection.delete_many({})
        collection = Connection('linked').collection('linkedcoursesstudents'
                                                     'linkedstudentscourses')
        collection.delete_many({})
        collection = Connection.get_collection(LinkedOrder)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedBuyer)
        collection.delete_many({})
        collection.delete_many({})
        collection = Connection.get_collection(LinkedCOrder)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedCBuyer)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedAccount)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedBalance)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedBomb)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedSoldier)
        collection.delete_many({})
        collection = Connection('linked').collection('linkedbombssoldiers'
                                                     'linkedsoldiersbombs')
        collection.delete_many({})
        collection = Connection.get_collection(LinkedCProfile)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedDUser)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedDProfile)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedCUser)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedNote)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedNotebook)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedRNote)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedRNotebook)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedCompany)
        collection.delete_many({})
        collection = Connection.get_collection(LinkedOwner)
        collection.delete_many({})
        collection = Connection('linked').collection('linkedcompaniesowners'
                                                     'linkedownerscompanies')
        collection.delete_many({})

    def test_object_can_be_removed_from_database(self):
        song = SimpleSong(name='Long', year=2020, artist='Thao')
        song.save()
        collection = Connection.get_collection(SimpleSong)
        self.assertEqual(collection.count_documents({}), 1)
        song.delete()
        self.assertEqual(collection.count_documents({}), 0)
        self.assertEqual(song.is_deleted, True)

    def test_1l_1f_list_can_be_nullified(self):
        author = LinkedAuthor(name='A', posts=[
            LinkedPost(title='P1', content='C1'),
            LinkedPost(title='P2', content='C2')])
        author.save()
        collection = Connection.get_collection(LinkedAuthor)
        self.assertEqual(collection.count_documents({}), 1)
        collection = Connection.get_collection(LinkedPost)
        self.assertEqual(collection.count_documents({}), 2)
        author.delete()
        for obj in collection.find():
            self.assertEqual(obj['authorId'], None)

    def test_1l_1f_instance_can_be_nullified(self):
        user = LinkedUser(name='crazy')
        profile = LinkedProfile(name='six')
        user.profile = profile
        user.save()
        collection = Connection.get_collection(LinkedUser)
        self.assertEqual(collection.count_documents({}), 1)
        collection = Connection.get_collection(LinkedProfile)
        self.assertEqual(collection.count_documents({}), 1)
        user.delete()
        for obj in collection.find():
            self.assertEqual(obj['userId'], None)

    def test_many_many_can_be_nullified(self):
        course1 = LinkedCourse(name='C1')
        course2 = LinkedCourse(name='C2')
        student1 = LinkedStudent(name='S1')
        student2 = LinkedStudent(name='S2')
        course1.students = [student1, student2]
        course2.students = [student1, student2]
        course1.save() # this triggers save also for course2
        collname = 'linkedcoursesstudentslinkedstudentscourses'
        collection = Connection('linked').collection(collname)
        self.assertEqual(collection.count_documents({}), 4)
        course1.delete()
        self.assertEqual(collection.count_documents({}), 2)

    def test_1f_1l_cascade_delete(self):
        account = LinkedAccount(name='A', balance=LinkedBalance(name='B'))
        account.save()
        collection = Connection.get_collection(LinkedAccount)
        self.assertEqual(collection.count_documents({}), 1)
        collection = Connection.get_collection(LinkedBalance)
        self.assertEqual(collection.count_documents({}), 1)
        account.delete()
        collection = Connection.get_collection(LinkedAccount)
        self.assertEqual(collection.count_documents({}), 0)
        collection = Connection.get_collection(LinkedBalance)
        self.assertEqual(collection.count_documents({}), 0)

    def test_1l_1f_cascade_delete(self):
        balance = LinkedBalance(name='A', account=LinkedAccount(name='B'))
        balance.save()
        collection = Connection.get_collection(LinkedAccount)
        self.assertEqual(collection.count_documents({}), 1)
        collection = Connection.get_collection(LinkedBalance)
        self.assertEqual(collection.count_documents({}), 1)
        balance.delete()
        collection = Connection.get_collection(LinkedAccount)
        self.assertEqual(collection.count_documents({}), 0)
        collection = Connection.get_collection(LinkedBalance)
        self.assertEqual(collection.count_documents({}), 0)

    def test_1_many_cascade_delete(self):
        buyer = LinkedBuyer(name='B')
        order1 = LinkedOrder(name='O1')
        order2 = LinkedOrder(name='O2')
        buyer.orders = [order1, order2]
        buyer.save()
        collection = Connection.get_collection(LinkedOrder)
        self.assertEqual(collection.count_documents({}), 2)
        buyer.delete()
        self.assertEqual(collection.count_documents({}), 0)

    def test_many_1_cascade_delete(self):
        buyer = LinkedCBuyer(name='B')
        order1 = LinkedCOrder(name='O1')
        order2 = LinkedCOrder(name='O2')
        buyer.orders = [order1, order2]
        buyer.save()
        collection = Connection.get_collection(LinkedCOrder)
        self.assertEqual(collection.count_documents({}), 2)
        order1.delete()
        self.assertEqual(collection.count_documents({}), 0)
        collection = Connection.get_collection(LinkedCBuyer)
        self.assertEqual(collection.count_documents({}), 0)

    def test_many_many_cascade_delete(self):
        soldier1 = LinkedSoldier(name='S1')
        soldier2 = LinkedSoldier(name='S2')
        bomb1 = LinkedBomb(name='B1')
        bomb2 = LinkedBomb(name='B2')
        soldier1.bombs = [bomb1, bomb2]
        soldier2.bombs = [bomb1, bomb2]
        soldier1.save()
        soldier2.save()
        collection = Connection.get_collection(LinkedSoldier)
        self.assertEqual(collection.count_documents({}), 2)
        collection = Connection.get_collection(LinkedBomb)
        self.assertEqual(collection.count_documents({}), 2)
        collection = Connection('linked').collection('linkedbombssoldiers'
                                                     'linkedsoldiersbombs')
        self.assertEqual(collection.count_documents({}), 4)
        soldier1.delete()
        collection = Connection.get_collection(LinkedSoldier)
        self.assertEqual(collection.count_documents({}), 0)
        collection = Connection.get_collection(LinkedBomb)
        self.assertEqual(collection.count_documents({}), 0)
        collection = Connection('linked').collection('linkedbombssoldiers'
                                                     'linkedsoldiersbombs')
        self.assertEqual(collection.count_documents({}), 0)

    def test_1l_1f_denies_deletion(self):
        user = LinkedDUser(name='crazy')
        profile = LinkedCProfile(name='six')
        user.profile = profile
        user.save()
        collection = Connection.get_collection(LinkedDUser)
        self.assertEqual(collection.count_documents({}), 1)
        collection = Connection.get_collection(LinkedCProfile)
        self.assertEqual(collection.count_documents({}), 1)
        self.assertRaises(DeletionDeniedException, user.delete)
        collection = Connection.get_collection(LinkedDUser)
        self.assertEqual(collection.count_documents({}), 1)
        collection = Connection.get_collection(LinkedCProfile)
        self.assertEqual(collection.count_documents({}), 1)

    def test_1l_1f_allows_deletion(self):
        user = LinkedDUser(name='crazy')
        user.save()
        collection = Connection.get_collection(LinkedDUser)
        self.assertEqual(collection.count_documents({}), 1)
        user.delete()
        collection = Connection.get_collection(LinkedDUser)
        self.assertEqual(collection.count_documents({}), 0)

    def test_1f_1l_denies_deletion(self):
        user = LinkedCUser(name='crazy')
        profile = LinkedDProfile(name='six')
        user.profile = profile
        user.save()
        collection = Connection.get_collection(LinkedCUser)
        self.assertEqual(collection.count_documents({}), 1)
        collection = Connection.get_collection(LinkedDProfile)
        self.assertEqual(collection.count_documents({}), 1)
        self.assertRaises(DeletionDeniedException, profile.delete)
        collection = Connection.get_collection(LinkedCUser)
        self.assertEqual(collection.count_documents({}), 1)
        collection = Connection.get_collection(LinkedDProfile)
        self.assertEqual(collection.count_documents({}), 1)

    def test1f_1l_allows_deletion(self):
        profile = LinkedDProfile(name='crazy')
        profile.save()
        collection = Connection.get_collection(LinkedDProfile)
        self.assertEqual(collection.count_documents({}), 1)
        profile.delete()
        collection = Connection.get_collection(LinkedDProfile)
        self.assertEqual(collection.count_documents({}), 0)

    def test_1_many_denies_deletion(self):
        notebook = LinkedNotebook(name='N',
                                  notes=[{'name': 'A'}, {'name': 'B'}])
        notebook.save()
        collection = Connection.get_collection(LinkedNotebook)
        self.assertEqual(collection.count_documents({}), 1)
        collection = Connection.get_collection(LinkedNote)
        self.assertEqual(collection.count_documents({}), 2)
        self.assertRaises(DeletionDeniedException, notebook.delete)
        collection = Connection.get_collection(LinkedNotebook)
        self.assertEqual(collection.count_documents({}), 1)
        collection = Connection.get_collection(LinkedNote)
        self.assertEqual(collection.count_documents({}), 2)

    def test_1_many_allows_deletion(self):
        notebook = LinkedNotebook(name='N', notes=[])
        notebook.save()
        collection = Connection.get_collection(LinkedNotebook)
        self.assertEqual(collection.count_documents({}), 1)
        notebook.delete()
        collection = Connection.get_collection(LinkedNotebook)
        self.assertEqual(collection.count_documents({}), 0)

    def test_many_1_denies_deletion(self):
        notebook = LinkedRNotebook(name='N',
                                  notes=[{'name': 'A'}, {'name': 'B'}])
        notebook.save()
        collection = Connection.get_collection(LinkedRNotebook)
        self.assertEqual(collection.count_documents({}), 1)
        collection = Connection.get_collection(LinkedRNote)
        self.assertEqual(collection.count_documents({}), 2)
        self.assertRaises(DeletionDeniedException, notebook.notes[0].delete)
        collection = Connection.get_collection(LinkedRNotebook)
        self.assertEqual(collection.count_documents({}), 1)
        collection = Connection.get_collection(LinkedRNote)
        self.assertEqual(collection.count_documents({}), 2)

    def test_many_1_allows_deletion(self):
        note = LinkedRNote(name='N')
        note.save()
        collection = Connection.get_collection(LinkedRNote)
        self.assertEqual(collection.count_documents({}), 1)
        note.delete()
        collection = Connection.get_collection(LinkedRNote)
        self.assertEqual(collection.count_documents({}), 0)

    def test_many_1_reverse_deny_has_nullify_effect(self):
        notebook = LinkedRNotebook(name='N',
                                   notes=[{'name': 'A'}, {'name': 'B'}])
        notebook.save()
        collection = Connection.get_collection(LinkedRNotebook)
        self.assertEqual(collection.count_documents({}), 1)
        collection = Connection.get_collection(LinkedRNote)
        self.assertEqual(collection.count_documents({}), 2)
        notebook.delete()
        collection = Connection.get_collection(LinkedRNotebook)
        self.assertEqual(collection.count_documents({}), 0)
        collection = Connection.get_collection(LinkedRNote)
        self.assertEqual(collection.count_documents({}), 2)
        for item in collection.find({}):
            self.assertEqual(item['notebookId'], None)

    def test_many_many_denies_deletion(self):
        c1 = LinkedCompany(name='C1')
        c2 = LinkedCompany(name='C2')
        o1 = LinkedOwner(name='O1')
        o2 = LinkedOwner(name='O2')
        c1.owners = [o1, o2]
        c2.owners = [o1, o2]
        o1.save()
        collection = Connection.get_collection(LinkedCompany)
        self.assertEqual(collection.count_documents({}), 2)
        collection = Connection.get_collection(LinkedOwner)
        self.assertEqual(collection.count_documents({}), 2)
        collection = Connection('linked').collection('linkedcompaniesowners'
                                                     'linkedownerscompanies')
        self.assertEqual(collection.count_documents({}), 4)
        self.assertRaises(DeletionDeniedException, o1.delete)
        collection = Connection.get_collection(LinkedCompany)
        self.assertEqual(collection.count_documents({}), 2)
        collection = Connection.get_collection(LinkedOwner)
        self.assertEqual(collection.count_documents({}), 2)
        collection = Connection('linked').collection('linkedcompaniesowners'
                                                     'linkedownerscompanies')
        self.assertEqual(collection.count_documents({}), 4)

    def test_many_many_allows_deletion(self):
        owner = LinkedOwner(name='OK')
        owner.save()
        collection = Connection.get_collection(LinkedOwner)
        self.assertEqual(collection.count_documents({}), 1)
        owner.delete()
        collection = Connection.get_collection(LinkedOwner)
        self.assertEqual(collection.count_documents({}), 0)

    def test_many_many_cascade_deny_triggers_partial_deletion(self):
        c1 = LinkedCompany(name='C1')
        c2 = LinkedCompany(name='C2')
        o1 = LinkedOwner(name='O1')
        o2 = LinkedOwner(name='O2')
        c1.owners = [o1, o2]
        c2.owners = [o1, o2]
        c1.save()
        c1.delete()
        collection = Connection.get_collection(LinkedCompany)
        self.assertEqual(collection.count_documents({}), 1)
        collection = Connection.get_collection(LinkedOwner)
        self.assertEqual(collection.count_documents({}), 2)
        collection = Connection('linked').collection('linkedcompaniesowners'
                                                     'linkedownerscompanies')
        self.assertEqual(collection.count_documents({}), 2)



from __future__ import annotations
from typing import cast, TYPE_CHECKING
from jsonclasses.jsonclass_field import JSONClassField
from jsonclasses.field_definition import FieldType, FieldStorage
from jsonclasses.types_resolver import TypesResolver
from inflection import camelize
from .connection import Connection
if TYPE_CHECKING:
    from .pymongo_object import PymongoObject


class Coder():

    def is_id_field(self, field: JSONClassField) -> bool:
        return field.definition.primary

    def is_instance_field(self, field: JSONClassField) -> bool:
        return field.definition.field_type == FieldType.INSTANCE

    def is_list_field(self, field: JSONClassField) -> bool:
        return field.definition.field_type == FieldType.LIST

    def is_list_instance_field(self, field: JSONClassField,
                                     cls: type[PymongoObject]) -> bool:
        if not self.is_list_field(field):
            return False
        t = TypesResolver().resolve_types(field.definition.raw_item_types,
                                          cls.definition.config)
        if t.definition.instance_types is not None:
            return True
        return False


    def is_foreign_key_storage(self, field: JSONClassField) -> bool:
        field_storage = field.definition.field_storage
        return field_storage == FieldStorage.FOREIGN_KEY

    def is_local_key_storage(self, field: JSONClassField) -> bool:
        field_storage = field.definition.field_storage
        return field_storage == FieldStorage.LOCAL_KEY

    def is_foreign_key_reference_field(self, field: JSONClassField) -> bool:
        return (self.is_instance_field(field) and
                self.is_foreign_key_storage(field))

    def is_foreign_keys_reference_field(self, field: JSONClassField) -> bool:
        return (self.is_list_field(field) and
                self.is_foreign_key_storage(field))

    def is_local_key_reference_field(self, field: JSONClassField) -> bool:
        return (self.is_instance_field(field) and
                self.is_local_key_storage(field))

    def is_local_keys_reference_field(self, field: JSONClassField) -> bool:
        return self.is_list_field(field) and self.is_local_key_storage(field)

    def is_join_table_field(self, field: JSONClassField) -> bool:
        return field.types.definition.use_join_table is True

    def list_instance_type(self,
                           field: JSONClassField,
                           sibling: type[PymongoObject]
                           ) -> type[PymongoObject]:
        from .pymongo_object import PymongoObject
        fd = field.types.definition
        item_types = TypesResolver().resolve_types(fd.raw_item_types,
                                                   sibling.definition.config)
        item_fd = item_types.definition
        return cast(type[PymongoObject], item_fd.instance_types)

    def join_table_name(self,
                        cls_a: type[PymongoObject],
                        field_a: str,
                        cls_b: type[PymongoObject],
                        field_b: str) -> str:
        connection = Connection.from_class(cls_a)
        cabase = connection.collection_from(cls_a).name
        cbbase = connection.collection_from(cls_b).name
        ca = cabase + camelize(field_a).lower()
        cb = cbbase + camelize(field_b).lower()
        return ca + cb if ca < cb else cb + ca



"""This module defines encoding and decoding context objects."""
from __future__ import annotations
from typing import NamedTuple, Any, Union, TYPE_CHECKING
from jsonclasses.types import Types
from jsonclasses.mark_graph import MarkGraph
if TYPE_CHECKING:
    from .pymongo_object import PymongoObject


class EncodingContext(NamedTuple):
    """Encoding context contains necessary information for encoding JSON Class
    objects into database writing commands.
    """
    value: Any
    types: Types
    keypath_root: str
    root: PymongoObject
    keypath_owner: str
    owner: PymongoObject
    keypath_parent: str
    parent: Union[list[Any], dict[str, Any], PymongoObject]
    mark_graph: MarkGraph

    def new(self, **kwargs):
        """Return a new encoding context by replacing provided values."""
        keys = kwargs.keys()
        return EncodingContext(
            value=kwargs['value'] if 'value' in keys else self.value,
            types=kwargs['types'] if 'types' in keys else self.types,
            keypath_root=(kwargs['keypath_root']
                          if 'keypath_root' in keys else self.keypath_root),
            root=kwargs['root'] if 'root' in keys else self.root,
            keypath_owner=(kwargs['keypath_owner']
                           if 'keypath_owner' in keys else self.keypath_owner),
            owner=kwargs['owner'] if 'owner' in keys else self.owner,
            keypath_parent=(kwargs['keypath_parent']
                            if 'keypath_parent' in keys
                            else self.keypath_parent),
            parent=kwargs['parent'] if 'parent' in keys else self.parent,
            mark_graph=(kwargs['mark_graph']
                        if 'mark_graph' in keys else self.mark_graph))



from __future__ import annotations
from jsonclasses import jsonclass
from jsonclasses_pymongo import pymongo, BaseObject


@pymongo
@jsonclass(class_graph='simple')
class SimpleSong(BaseObject):
    name: str
    year: int
    artist: str



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(name='jsonclasses-pymongo',
      version='2.3.0',
      description='jsonclasses integration with pymongo',
      long_description=README,
      long_description_content_type="text/markdown",
      author='Fillmula Inc.',
      author_email='wiosoftvictor@163.com',
      license='MIT',
      packages=find_packages(exclude=("tests")),
      package_data={'jsonclasses_pymongo': ['py.typed']},
      zip_safe=False,
      url='https://github.com/fillmula/jsonclasses-pymongo',
      include_package_data=True,
      python_requires='>=3.9')



from __future__ import annotations
from jsonclasses import jsonclass, types
from jsonclasses_pymongo import pymongo, BaseObject


@pymongo(collection_name='simplepeople')
@jsonclass(class_graph='simple')
class SimplePerson(BaseObject):
    name: str = types.str.unique.required
